# -*- coding: utf-8 -*-

"""Module that exposes the routines and utilities making up SPYDCMTK
"""

import os
import sys
import base64
import argparse
import numpy as np
import shutil
import pydicom as dicom

# Local imports 
import spydcmtk.dcmTools as dcmTools
import spydcmtk.dcmTK as dcmTK
from spydcmtk.spydcm_config import SpydcmTK_config


class INTERACTIVE():
    def __init__(self, study, outputPath) -> None:
        self.study = study
        self.outputPath = outputPath
        #
        self.options = {
            '1': ("Data summary", self.dataSummary),
            '2': ("Build VTS", self.buildVTS),
            '3': ("Build FDQ", self.buildFDQ),
            '4': ("Build overview image", self.buildOverviewImage),
            'q': ("Quit", self.quit)
        }

    def displayMenu(self):
        print("\nSelect an option:")
        for key, (description, _) in self.options.items():
            print(f"{key}: {description}")

    def getUserInput(self, question="your choice"):
        return input(f"Enter {question}: ")

    def run(self):
        while True:
            self.displayMenu()
            choice = self.getUserInput()
            if choice in self.options:
                self.options[choice][1]()  # Execute the corresponding function
            else:
                print("Invalid option. Please try again.")

    def dataSummary(self):
        print(self.study.getStudySummary())

    def buildVTS(self):
        self.dataSummary()
        seNum = self.getUserInput("series number")
        try: 
            seNum = int(seNum)
            dcmSeries = self.study.getSeriesByID(seNum)
            outputFilename = self.getUserInput("file name (vts)")
            outputpath = os.path.join(self.outputPath, outputFilename)
            dcmSeries.writeToVTS(outputpath)
        except ValueError:
            print("Invalid option. Please try again")

    def buildFDQ(self):
        self.dataSummary()
        seNum_ = self.getUserInput("series numbers for FDQ (MAX, PX, PY, PZ):")
        try: 
            seNum_4 = seNum_.strip().split(' ')
            seNum_4 = [int(i) for i in seNum_4]
            outputFilename = self.getUserInput("file name (pvd)")
            outputpath = os.path.join(self.outputPath, outputFilename)
            self.study.writeFDQ(seNum_4, outputpath)
        except ValueError:
            print("Invalid option. Please try again")

    def buildOverviewImage(self):
        self.dataSummary()
        seNum = self.getUserInput("series number")
        outputFilename = self.getUserInput("file name (png) - leave blank to show")
        if outputFilename == '':
            outputpath = None
        else:
            outputpath = os.path.join(self.outputPath, outputFilename)
        self.study.getSeriesByID(seNum).buildOverviewImage(outputpath)

    def quit(self):
        print("Exiting the menu.")
        exit()  


def writeDirectoryToNII(dcmDir, outputPath, fileName):
    """ Write a directory of dicom files to a Nifti (`*.nii.gz`) file. 
        Uses dcm2nii so MUST have dcm2nii installed and within path.

    Args:
        dcmDir (str): Path to directory containing dicom files
        outputPath (str): Path to output directory where to save nifti
        fileName (str): Name of output nii.gz file (will rename nii.gz output from dcm2nii)
    """
    return dcmTools.writeDirectoryToNII(dcmDir, outputPath, fileName)


def buildTableOfDicomParamsForManuscript(topLevelDirectoryList, outputCSVPath, seriesDescriptionIdentifier=None, ONE_FILE_PER_DIR=True):
    """ Build a simple table of dicom parameters that would be suitable for input into a scientific manuscript.
        Output is comma delimited multiline string of Tr, Te (etc) stats. 

    Args:
        topLevelDirectoryList (list): list of str that are paths to directories containing dicom files
        outputCSVPath (str): path to outputCSV
        seriesDescriptionIdentifier (str, optional): A substring of SeriesDescription tag used to select certian series only. Defaults to None.
        ONE_FILE_PER_DIR (bool, optional): Set false to read all dicoms, can set True for faster reading if dicoms organised. Default: True
    """
    dfData = []
    for inspectDir in topLevelDirectoryList:
        try:
            ssL = dcmTK.ListOfDicomStudies.setFromDirectory(inspectDir, OVERVIEW=True, HIDE_PROGRESSBAR=True, ONE_FILE_PER_DIR=ONE_FILE_PER_DIR)
        except IndexError: # no dicoms found 
            continue
        for ss in ssL:
            if seriesDescriptionIdentifier is not None:
                matchingSeries = ss.getSeriesMatchingDescription([seriesDescriptionIdentifier], RETURN_SERIES_OBJ=True)
            else:
                matchingSeries = ss
            if matchingSeries is not None:
                for iSeries in matchingSeries:
                    resDict = iSeries.getSeriesInfoDict(EXTRA_TAGS=SpydcmTK_config.MANUSCRIPT_TABLE_EXTRA_TAG_LIST)
                    resDict["Identifier"] = dcmTools.getDicomFileIdentifierStr(iSeries[0])
                    dfData.append(resDict)
    stats = {}
    dfData = sorted(dfData, key=lambda i:i["Identifier"])
    tagList = sorted(dfData[0].keys())
    strOut = ','+','.join(tagList) + '\n'
    for row in dfData:
        strOut += ','
        for i in tagList:
            strOut += f'{row[i]},'
            stats.setdefault(i, []).append(row[i])
        strOut += '\n'
    strOut += '\n\nSTATS:\n'
    for label, myFunc in zip(['Mean', 'Standard Deviation', 'Median', 'Min', 'Max'], 
                             [np.mean, np.std, np.median, np.min, np.max]):
        for k1 in range(len(tagList)):
            if k1 == 0:
                strOut += f'{label},'
            try:
                strOut += f'{myFunc(stats[tagList[k1]])},'
            except:
                strOut += 'NAN,'
        strOut += '\n'
    with open(outputCSVPath, 'w') as fid:
        fid.write(strOut)
    return outputCSVPath


def getAllDirsUnderRootWithDicoms(rootDir, QUIET=True, FORCE_READ=False):
    fullDirsWithDicoms = []
    for root, _, files in os.walk(rootDir):
        for iFile in files:
            thisFile = os.path.join(root, iFile)
            try:
                dicom.dcmread(thisFile, stop_before_pixels=True, defer_size=16, force=FORCE_READ) # will error if not dicom
                if not QUIET:
                    print('OK: %s'%(thisFile))
                fullDirsWithDicoms.append(root)
                break
            except dicom.filereader.InvalidDicomError:
                if not QUIET:
                    print('FAIL: %s'%(thisFile))
                continue
    return fullDirsWithDicoms


def anonymiseInPlace(dicomDirectory, anonName=None, anonID="", QUIET=False):
    if anonName is None:
        ds = returnFirstDicomFound(dicomDirectory)
        if ds is None:
            return # Nothing to anonymise... 
        anonName = ""
    #
    tmpDir = dicomDirectory+".TEMP.WORKING"
    os.rename(dicomDirectory, tmpDir)
    dcmTools.streamDicoms(tmpDir, dicomDirectory, anonName=anonName, anonID=anonID, HIDE_PROGRESSBAR=QUIET)
    shutil.rmtree(tmpDir)


def returnFirstDicomFound(rootDir, FILE_NAME_ONLY=False, MatchingTag_dict=None):
    """
    Search recursively for first dicom file under root and return.
    If have dicoms in nice folder structure then this can be a fast way to find, e.g. all series with protocol X

    :param rootDir: directory on filesystem
    :param FILE_NAME_ONLY: If true will return the file name [Default False]
    :param MatchingTag_dict: If given then will only consider dicoms where tag(key) matches value given [Default None]
    :return: pydicom dataset<without pixel data> or fileName<str>
    """
    for root, _, files in os.walk(rootDir):
        for iFile in files:
            if 'dicomdir' in iFile.lower():
                continue
            thisFile = os.path.join(root, iFile)
            try:
                dataset = dicom.dcmread(thisFile, stop_before_pixels=True)
                if MatchingTag_dict is not None:
                    tf = []
                    for iKey in MatchingTag_dict.keys():
                        tf.append(str(dataset[iKey].value) == str(MatchingTag_dict[iKey]))
                    if not all(tf):
                        continue
                if FILE_NAME_ONLY:
                    return thisFile
                else:
                    return dataset
            except dicom.filereader.InvalidDicomError:
                continue
    return None

def getTag(pathToDicoms, tagName):
    """Convienience function to find first dicom and then get tag value

    Args:
        pathToDicoms (str): path to dicoms
        tagName (str): dicom tag name
    """
    ds = returnFirstDicomFound(pathToDicoms)
    return ds[tagName].value, ds.filename

def directoryToVTI(dcmDirectory, outputFolder, 
                   outputNamingTags=SpydcmTK_config.VTI_NAMING_TAG_LIST, 
                   QUITE=True, FORCE=False, TRUE_ORIENTATION=False):
    """Convert directory of dicoms to VTI files (one vti per series)
        Naming built from dicom tags: 

    Args:
        dcmDirectory (str): Directory containing dicom files
        outputFolder (str): Directory where vti output files to be written
        outputNamingTags (tuple, optional): Dicom tags used to generate vti file name. 
                        Defaults to ('PatientName', 'SeriesNumber', 'SeriesDescription')
        QUITE (bool, optional): Suppress output information. Defaults to True.
        FORCE (bool, optional): Set True to overwrite already present files. Defaults to False.

    Returns:
        list: List of output file names written
    """
    ListDicomStudies = dcmTK.ListOfDicomStudies.setFromInput(dcmDirectory, HIDE_PROGRESSBAR=QUITE, FORCE_READ=FORCE, OVERVIEW=False) 
    return _listDicomStudiesToVTI(ListDicomStudies, outputFolder=outputFolder, outputNamingTags=outputNamingTags, TRUE_ORIENTATION=TRUE_ORIENTATION)


def directoryToVTS(dcmDirectory, outputFolder,
                   outputNamingTags=SpydcmTK_config.VTI_NAMING_TAG_LIST, 
                   QUITE=True, FORCE=False):
    """Convert directory of dicoms to VTS files (Good for cine etc)
        Naming built from dicom tags: 

    Args:
        dcmDirectory (str): Directory containing dicom files
        outputFolder (str): Directory where vti output files to be written
        outputNamingTags (tuple, optional): Dicom tags used to generate vti file name. 
                        Defaults to ('PatientName', 'SeriesNumber', 'SeriesDescription')
        QUITE (bool, optional): Suppress output information. Defaults to True.
        FORCE (bool, optional): Set True to overwrite already present files. Defaults to False.

    Returns:
        list: List of output file names written
    """
    ListDicomStudies = dcmTK.ListOfDicomStudies.setFromInput(dcmDirectory, HIDE_PROGRESSBAR=QUITE, FORCE_READ=FORCE, OVERVIEW=False) 
    return _listDicomStudiesToVTI(ListDicomStudies, outputFolder=outputFolder, outputNamingTags=outputNamingTags, VTS=True)


def _listDicomStudiesToVTI(ListDicomStudies, outputFolder, outputNamingTags=SpydcmTK_config.VTI_NAMING_TAG_LIST, QUIET=True, TRUE_ORIENTATION=False, VTS=False):
    outputFiles = []
    for iDS in ListDicomStudies:
        for iSeries in iDS:
            if VTS:
                # print(f"DICOMS TO VTS IS NOT IMPLEMENTED YET")
                fOut = iSeries.writeToVTS(outputPath=outputFolder, outputNamingTags=outputNamingTags)
            else:
                fOut = iSeries.writeToVTI(outputPath=outputFolder, outputNamingTags=outputNamingTags, TRUE_ORIENTATION=TRUE_ORIENTATION)
            outputFiles.append(fOut)
            if not QUIET:
                print(f'Written {fOut}')
    return outputFiles

# =========================================================================
# =========================================================================
## DATA TO HTML
# =========================================================================
def convertInputsToHTML(listOfFilePaths, outputFile=None, glanceHtml=None, QUIET=False, DEBUG=False):
    """Convert inputs to HTML viewable via ParaViewGlance (3D volumes and/or surface polymeshes)
    
    Keyword arguments:
    listOfFilePaths             -- May be nii, vtk or dicoms
    outputFile (optional)       -- Full path to outputfile, if not given then use first input path. Defaults to None.
    glanceHtml (optional)       -- Full path to glanceHtml template file, if not given then use default. Defaults to None.
    QUIET (optional)            -- Set True to suppress output. Defaults to False.
    DEBUG (optional)            -- Set True to prevent cleanup of intermediary files. Defaults to False.
    Return: path to html file written
    """
    
    CLEAN_UP_LIST = []
    FILE_TO_VTK_LIST = []
    ## --- Check inputs ---
    # Glance html
    if glanceHtml is None:
        thisDir = os.path.split(os.path.realpath(__file__))[0]
        glanceHtml = os.path.join(thisDir, 'ParaViewGlance.html')
        if not QUIET:
            print('Using ParaView glance file: %s'%(glanceHtml))
    if not os.path.isfile(glanceHtml):
        raise ValueError('%s does not exist'%(glanceHtml))
    if type(listOfFilePaths) != list:
        listOfFilePaths = [listOfFilePaths]
    ## --- Output file ---
    if outputFile is None:
        outputDir, fName = os.path.split(listOfFilePaths[0])
        fNameOut = os.path.splitext(fName)[0]+'.html'
        outputFile = os.path.join(outputDir, fNameOut)
    elif os.path.isdir(outputFile):
        outputDir = outputFile
        _, fName = os.path.split(listOfFilePaths[0])
        fNameOut = os.path.splitext(fName)[0]+'.html'
        outputFile = os.path.join(outputDir, fNameOut)
    else:
        outputDir, fNameOut = os.path.split(outputFile)

    ## --- VTK Objs / file paths ---
    for iPath in listOfFilePaths:
        if os.path.isfile(iPath):
            if iPath.endswith('nii') or iPath.endswith('nii.gz') :
                iPath = dcmTK.dcmVTKTK.fIO.readNifti(iPath)
                CLEAN_UP_LIST.append(iPath)
            FILE_TO_VTK_LIST.append(iPath)
        else:
            if os.path.isdir(iPath): # If path to dicoms
                dcmToVTKPath = directoryToVTI(iPath, outputDir, TRUE_ORIENTATION=False)
                for ifile in dcmToVTKPath:
                    if ifile.endswith('.pvd'):
                        FILE_TO_VTK_LIST += list(dcmTK.dcmVTKTK.fIO.readPVDFileName(ifile).values())
                    else:
                        FILE_TO_VTK_LIST.append(ifile)
                CLEAN_UP_LIST += dcmToVTKPath
            else:
                raise ValueError('%s does not exist' % (iPath))

    ## --- Build HTML Recursivly ---
    if not QUIET:
        print('Writing %s from base html %s, using:'%(outputFile, glanceHtml))
        for iFile in listOfFilePaths:
            print('    %s'%(iFile))

    for k1, iFile in enumerate(FILE_TO_VTK_LIST):
        outputTemp = outputFile[:-5]+'_TEMP%d.html'%(k1)
        glanceHtml = vtkToHTML(iFile, glanceHtml, outputTemp)
        CLEAN_UP_LIST.append(outputTemp)
    os.rename(CLEAN_UP_LIST.pop(), outputFile)

    ## --- Clean up --- // Skipped if in DEBUG mode
    if not DEBUG:
        if not QUIET:
            print('Cleaning up:', str(CLEAN_UP_LIST))
        for ifile in CLEAN_UP_LIST:
            if iFile.endswith('.pvd'):
                dcmTK.dcmVTKTK.fIO.deleteFilesByPVD(iFile)
            else:
                os.unlink(ifile)

    return outputFile


def vtkToHTML(vtkDataPath, glanceHtml, outputFile):
    # Extract data as base64
    with open(vtkDataPath, "rb") as data:
        dataContent = data.read()
        base64Content = base64.b64encode(dataContent)
        base64Content = base64Content.decode().replace("\n", "")
    # Create new output file
    with open(glanceHtml, mode="r", encoding="utf-8") as srcHtml:
        with open(outputFile, mode="w", encoding="utf-8") as dstHtml:
            for line in srcHtml:
                if "</body>" in line:
                    dstHtml.write("<script>\n")
                    dstHtml.write('var contentToLoad = "%s";\n\n' % base64Content)
                    dstHtml.write(
                        'Glance.importBase64Dataset("%s" , contentToLoad, glanceInstance.proxyManager);\n'
                        % os.path.basename(vtkDataPath)
                    )
                    dstHtml.write("glanceInstance.showApp();\n")
                    dstHtml.write("</script>\n")
                dstHtml.write(line)
    return outputFile


### ====================================================================================================================
##          RUN VIA MAIN
### ====================================================================================================================
# Override error to show help on argparse error (missing required argument etc)
class MyParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)
##  ========= INSPECTION / VERIFICATION =========
def inspect(dStudies, FULL):
    frmtStr = dStudies.getSummaryString(FULL)
    print(frmtStr)

def checkArgs(args):
    allActionArgs = [args.nii, 
                args.inspect, 
                args.inspectFull, 
                args.inspectQuick,
                args.outputFolder is not None,
                args.vti,
                args.INTERACTIVE]
    return any(allActionArgs)

##  ========= RUN ACTIONS =========
def _runActions(args, ap):

    ####
    if args.dcmdump:
        ds = returnFirstDicomFound(args.inputPath, FILE_NAME_ONLY=False)
        print(ds)
    if args.tagValue is not None:
        val, fileName = getTag(args.inputPath, args.tagValue)
        print(f"{args.tagValue} = {val} (found at {fileName})")
    elif args.msTable:
        outputCSV = args.outputFolder if args.outputFolder.endswith('.csv') else os.path.join(args.outputFolder, 'ms.csv')
        fOut = buildTableOfDicomParamsForManuscript([args.inputPath], 
                                                    outputCSVPath= outputCSV,
                                                    seriesDescriptionIdentifier=None,
                                                    ONE_FILE_PER_DIR=False)
        print(f"Written Manuscript table like csv to {fOut}")
    else:
        # check arguments to avoid reading all dicoms and then doing nothing...
        if not checkArgs(args):
            ap.exit(0, f'No action given. Exiting SPYDCMTK without action\n')
        if args.STREAM:
            dcmTools.streamDicoms(args.inputPath, args.outputFolder, FORCE_READ=args.FORCE, HIDE_PROGRESSBAR=args.QUIET, SAFE_NAMING=args.SAFE)
            return 0
        ##
        ## NOW READING DICOMS
        try:
            onlyOverview = args.inspect or args.inspectFull or args.INTERACTIVE
            oneFilePerDir = args.inspectQuick or args.INTERACTIVE or args.seNumber or args.filter
            # If read one file per dir then will read all dicoms upon write out (or conversion). 
            # Only issue is if dicoms not well organised - TODO - NOTE this somewhere... Maybe bettter ad an option for force read all by user. 
            if not args.QUIET:
                print(f"READING...")
            ListDicomStudies = dcmTK.ListOfDicomStudies.setFromInput(args.inputPath, 
                                                                     ONE_FILE_PER_DIR=oneFilePerDir,
                                                                     HIDE_PROGRESSBAR=args.QUIET, 
                                                                     FORCE_READ=args.FORCE, 
                                                                     OVERVIEW=onlyOverview) 
            if args.SAFE:
                ListDicomStudies.setSafeNameMode()

            if args.seNumber is not None:
                if len(ListDicomStudies) > 1:
                    ListDicomStudies = ListDicomStudies.filterByTag('SeriesNumber', args.seNumber)
                else:
                    newListOfDicomStudies = []
                    for iStudy in ListDicomStudies: 
                        newListOfDicomStudies.append(iStudy.filterByTag('SeriesNumber', args.seNumber))
                    ListDicomStudies = dcmTK.ListOfDicomStudies(newListOfDicomStudies)

            if args.filter is not None:
                if len(ListDicomStudies) > 1:
                    ListDicomStudies = ListDicomStudies.filterByTag(args.filter[0], args.filter[1])
                else:
                    newListOfDicomStudies = []
                    for iStudy in ListDicomStudies: 
                        newListOfDicomStudies.append(iStudy.filterByTag(args.filter[0], args.filter[1]))
                    ListDicomStudies = dcmTK.ListOfDicomStudies(newListOfDicomStudies)

        except IOError as e:
            ap.exit(1, f'Error reading {args.inputPath}.\n    {e}')
            # Let IOERROR play out here is not correct input
        ##
        ## NOW ACTIONS
        if args.inspect or args.inspectFull or args.inspectQuick:
            for iStudy in ListDicomStudies:
                print(iStudy.getTopDir())
                print(iStudy.getStudySummary(args.inspectFull))
                print('\n')

        elif args.vti2dcm is not None:
            if not os.path.isfile(args.vti2dcm):
                ap.exit(1, f'ERROR: VTI file {args.vti2dcm} not found')
            if not args.vti2dcm.lower().endswith('.vti'):
                ap.exit(1, f'ERROR: VTI file {args.vti2dcm} is not a valid VTI file')
            vtiObj = dcmTK.dcmVTKTK.fIO.readVTKFile(args.vti2dcm)
            series = ListDicomStudies[0][0]
            series.sortBySlice_InstanceNumber()
            dcmDirOut = dcmTK.writeVTIToDicoms(vtiObj, series[0], outputDir=args.outputFolder)
            if not args.QUIET:
                print(f'Written {dcmDirOut}')
        # elif args.nii2dcm is not None:
        #     dcmTK.dcmVTKTK.fIO.readNifti(args.nii2dcm)
        elif args.INTERACTIVE:
            # TODO - check if multiple studies
            INTER = INTERACTIVE(ListDicomStudies[0], outputPath=args.outputFolder)
            INTER.run()

        else:
            if args.outputFolder is None:
                print(f'WARNING: outputFolder not given - setting to inputFolder')
                args.outputFolder = os.path.split(args.inputPath)[0]
            if args.nii:
                for iDS in ListDicomStudies:
                    for iSeries in iDS:
                        fOut = iSeries.writeToNII(outputPath=args.outputFolder)
                        if not args.QUIET:
                            print(f'Written {fOut}')
            elif args.vti:
                _listDicomStudiesToVTI(ListDicomStudies=ListDicomStudies, outputFolder=args.outputFolder, QUIET=args.QUIET, TRUE_ORIENTATION=args.TRUE_VTI_ORIENTATION)
            elif args.vts:
                _listDicomStudiesToVTI(ListDicomStudies=ListDicomStudies, outputFolder=args.outputFolder, QUIET=args.QUIET, VTS=True)
            elif args.html:
                for iDS in ListDicomStudies:
                    for iSeries in iDS:
                        fOut = convertInputsToHTML([iSeries.getRootDir()], args.outputFolder)
                        if not args.QUIET:
                            print(f'Written {fOut}')
            elif args.outputFolder is not None:
                if args.anonName is not None:
                    if not args.QUIET:
                        print(f"ANONYMISING...")
                    ListDicomStudies.anonymise(anonName=args.anonName, 
                                               anonPatientID=args.anonID, 
                                               removePrivateTags=args.REMOVE_PRIVATE_TAGS)
                    ListDicomStudies.resetUIDs()
                if not args.QUIET:
                    print(f"WRITTING...")
                outDirList = ListDicomStudies.writeToOrganisedFileStructure(args.outputFolder)
                if len(outDirList) > 0:
                    allDirsPresent = all([os.path.isdir(i) for i in outDirList])
                    res = 0 if allDirsPresent else 1
                    ap.exit(res, f'Transfer and sort from {args.inputPath} to {args.outputFolder} COMPLETE\n')
                else:
                    ap.exit(1, f'No dicoms written out for given conditions from {args.inputPath}\n')
        ##

### ====================================================================================================================
### ====================================================================================================================
# S T A R T
#
def main():
    # --------------------------------------------------------------------------
    #  ARGUMENT PARSING
    # --------------------------------------------------------------------------
    ap = MyParser(description='Simple Python Dicom Toolkit - spydcmtk')

    ap.add_argument('-i', dest='inputPath', help='Path to find dicoms (file or directory or tar or tar.gz or zip)', type=str, default=None)
    ap.add_argument('-o', dest='outputFolder', help='Path for output - if set then will organise dicoms into this folder', type=str, default=None)

    ap.add_argument('-a', dest='anonName',
        help='anonymous name [optional - if not given, then not anoymised]', type=str, default=None)
    ap.add_argument('-aid', dest='anonID',
        help='anonymous ID [optional - only used if anonName is given, default=""]', type=str, default='')
    ap.add_argument('-RemovePrivateTags', dest='REMOVE_PRIVATE_TAGS',
        help='set to remove private tags during anonymisation [optional - only used if anonName is given, default="False"]', action='store_true')
    ap.add_argument('-filter', dest='filter',
        help='Will filter dicoms based on tag name and value. Tag name and value required. If input is multiple studies then act on each.', nargs=2, default=None)
    ap.add_argument('-seNumber', dest='seNumber',
        help='Will filter dicoms based on series number. Same as -filter SeriesNumber #', type=str, default=None)
    ap.add_argument('-inspect', dest='inspect',
        help='Will output a summary of dicoms to the terminal', action='store_true')
    ap.add_argument('-inspectFull', dest='inspectFull',
        help='Will output a full summary of dicoms to the terminal', action='store_true')
    ap.add_argument('-inspectQuick', dest='inspectQuick',
        help='Run inspection - expect organised file structure (read only one file per directory)', action='store_true')
    ap.add_argument('-msTable', dest='msTable',
        help='Will output a csv to outputFolder with tags info suitable for building manuscript style table', action='store_true')
    ap.add_argument('-dcmdump', dest='dcmdump',
        help='Will output a dump of all dicom tags to the terminal (from first found dicom)', action='store_true')
    ap.add_argument('-tag', dest='tagValue',
        help='Get tag value for first dicom found under input', type=str, default=None)
    ap.add_argument('-nii', dest='nii',
        help='Will convert each series to nii.gz. Naming: {PName}_{SE#}_{SEDesc}.nii.gz', action='store_true')
    ap.add_argument('-vti', dest='vti',
        help='Will convert each series to vti. Naming: {PName}_{SE#}_{SEDesc}.vti', action='store_true')
    ap.add_argument('-TRUE_VTI_ORIENTATION', dest='TRUE_VTI_ORIENTATION', help='Will resample vti data at true location (output different dimensiuons)', action='store_true')
    ap.add_argument('-vts', dest='vts',
        help='Will convert each series to vts. Naming: {PName}_{SE#}_{SEDesc}.vts', action='store_true')
    ap.add_argument('-html', dest='html',
        help='Will convert each series to html file for web viewing. Naming: outputfolder argument', action='store_true')
    #
    ap.add_argument('-vti2dcm', dest='vti2dcm',
        help='Will convert VTI to dicom series. Pass reference dicoms as input.', type=str, default=None)
    # ap.add_argument('-nii2dcm', dest='nii2dcm',
    #     help='Will convert NII to dicom series. Pass reference dicoms as input.', type=str, default=None)
    #
    ap.add_argument('-I', dest='INTERACTIVE', help='Will read input and launch interactive mode', action='store_true')
    #
    ap.add_argument('-config', dest='configFile', help='Path to configuration file to use.', type=str, default=None)
    # -- program behaviour guidence -- #
    ap.add_argument('-SAFE', dest='SAFE', help='Safe naming - uses UIDs for naming to avoid potential conflicts.\n\t'+\
                        'Note, normal behaviour is to check for safe naming but certain conditions may require this.',
                            action='store_true')
    ap.add_argument('-FORCE', dest='FORCE', help='force reading even if not standard dicom (needed if dicom files missing header meta)',
                            action='store_true')
    ap.add_argument('-QUIET', dest='QUIET', help='Suppress progress bars and information output to terminal',
                            action='store_true')
    ap.add_argument('-INFO', dest='INFO', help='Provide setup (configuration) info and exit.',
                            action='store_true')
    ap.add_argument('-STREAM', dest='STREAM', help='To organise dicoms rapidly without any quality checking',
                            action='store_true')
    ##

    arguments = ap.parse_args()
    SpydcmTK_config.runconfigParser(arguments.configFile)
    if arguments.INFO:
        SpydcmTK_config.printInfo()
        sys.exit(1)

    if arguments.inputPath is not None:
        arguments.inputPath = os.path.abspath(arguments.inputPath)
        if not (os.path.isdir(arguments.inputPath) or os.path.isfile(arguments.inputPath)):
            print(f'## ERROR : Can not find input {arguments.inputPath}')
            print('EXITING')
            sys.exit(1)
        if not arguments.QUIET:
            print(f'Running SPYDCMTK with input {arguments.inputPath}')
    else:
        print(f'## ERROR : inputPath (-i option) must be provided. ')
        print('EXITING')
        sys.exit(1)

    ## -------------
    if arguments.outputFolder is not None:
        arguments.outputFolder = os.path.abspath(arguments.outputFolder)
        # Check if input is subdirectory of output
        if os.path.commonpath([arguments.inputPath]) == os.path.commonpath([arguments.inputPath, arguments.outputFolder]):
            print(f'## ERROR: Input directory ({arguments.inputPath}) cannot be a subdirectory of output directory ({arguments.outputFolder})')
            print('This would cause recursive copying and potential data corruption.')
            print('EXITING')
            sys.exit(1)

    _runActions(arguments, ap)


if __name__ == '__main__':

    main()


"""
Created on MArch 2023 (rewrite from old module - remove reliance on VTKDICOM)

@author: fraser

Dicom to VTK conversion toolkit

"""

import os
import numpy as np
from typing import Optional, Dict, List, Any
import pydicom as dicom
from pydicom.sr.codedict import codes
from pydicom.uid import generate_uid
from highdicom.seg.content import SegmentDescription
from highdicom.seg.enum import SegmentAlgorithmTypeValues, SegmentationTypeValues
from highdicom.content import AlgorithmIdentificationSequence
from highdicom.seg.sop import Segmentation

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET

import vtk
from vtk.util import numpy_support # type: ignore

import spydcmtk.dcmTools as dcmTools
from ngawari import fIO
from ngawari import ftk
from ngawari import vtkfilters
from multiprocessing import Pool


mm_to_m = 0.001
cm_to_m = 0.01
ms_to_s = 0.001
m_to_mm = 1000.0

# =========================================================================
## PATIENT MATRIX HELPER
# =========================================================================
class PatientMeta:
    """A class that manages spatial / geometric information for DICOM and VTK conversion
    
    _meta keys:
    'ImagePositionPatient', 'PixelSpacing', 'ImageOrientationPatient', 'SliceVector', 'SliceThickness', 'SliceLocation0', 'SpacingBetweenSlices', 'Dimensions'
    """

    def __init__(self):
        self.units = "SI"
        # Minimal defaults
        self._meta = {
            'ImageOrientationPatient': [1.0, 0.0, 0.0, 0.0, 1.0, 0.0],
            'PixelSpacing': [1.0, 1.0],
            'SliceThickness': 1.0,
            'ImagePositionPatient': [0.0, 0.0, 0.0],
            'SliceLocation0': 0.0
            }
        self._matrix = None

    def __str__(self):
        return str(self._meta)

    # Properties
    @property
    def PixelSpacing(self):
        return self._meta['PixelSpacing']
    
    @property
    def ImagePositionPatient(self):
        return self._meta['ImagePositionPatient']
    
    @property
    def ImageOrientationPatient(self):
        return self._meta['ImageOrientationPatient']
    
    @property
    def SliceVector(self):
        try:
            return self._meta['SliceVector']
        except KeyError:
            return np.cross(self._meta['ImageOrientationPatient'][:3], self._meta['ImageOrientationPatient'][3:6])
    
    @property
    def SpacingBetweenSlices(self):
        try:
            return self._meta['SpacingBetweenSlices']
        except KeyError:
            return self._meta['SliceThickness']
    
    @property
    def SliceThickness(self):
        try:
            return self._meta['SliceThickness']
        except KeyError:
            return self._meta['SpacingBetweenSlices']
    
    @property
    def SliceLocation0(self):
        try:
            return self._meta['SliceLocation0']
        except KeyError:
            return 0.0
    
    @property
    def Dimensions(self):
        return self._meta['Dimensions']

    @property
    def Origin(self):
        return self.ImagePositionPatient
    
    @property
    def Spacing(self):
        if 'SpacingBetweenSlices' in self._meta:
            return self._meta['PixelSpacing'][0], self._meta['PixelSpacing'][1], self._meta['SpacingBetweenSlices']
        else:
            return self._meta['PixelSpacing'][0], self._meta['PixelSpacing'][1], self._meta['SliceThickness']
    
    @property
    def PatientPosition(self):
        try:
            return self._meta['PatientPosition']
        except KeyError:
            return 'HFS'

    @property
    def Times(self):
        try:
            return self._meta['Times']
        except KeyError:
            return [0.0]

    # ------------------------------------------------------------------------------------------------------------------------------
    def initFromDictionary(self, metaDict):
        # Force required keys
        if 'PixelSpacing' not in metaDict:
            if 'Spacing' not in metaDict:
                metaDict['Spacing'] = [1.0, 1.0, 1.0]
        if 'ImagePositionPatient' not in metaDict:
            if 'Origin' not in metaDict:
                metaDict['Origin'] = [0.0, 0.0, 0.0]
        if 'ImageOrientationPatient' not in metaDict:
            metaDict['ImageOrientationPatient'] = [1.0, 0.0, 0.0, 0.0, 1.0, 0.0]
        self._meta.update(metaDict)
        self.__metaVTI2DCMConversion()
        self._updateMatrix()

    def __metaVTI2DCMConversion(self):
        if 'Spacing' in self._meta:
            self._meta['PixelSpacing'] = [self._meta['Spacing'][0], self._meta['Spacing'][1]]
            if 'SliceThickness' not in self._meta:
                self._meta['SliceThickness'] = self._meta['Spacing'][2]
        if 'SpacingBetweenSlices' not in self._meta:
            self._meta['SpacingBetweenSlices'] = self._meta['SliceThickness']
        if 'Origin' in self._meta:
            self._meta['ImagePositionPatient'] = [self._meta['Origin'][0], self._meta['Origin'][1], self._meta['Origin'][2]]

    def initFromDicomSeries(self, dicomSeries):
        I,J,K = int(dicomSeries.getTag('Rows')), int(dicomSeries.getTag('Columns')), int(dicomSeries.getNumberOfSlicesPerVolume())
        dicomSeries.sortBySlice_InstanceNumber()
        N = dicomSeries.getNumberOfTimeSteps()
        A = np.zeros((I, J, K, N))
        c0 = 0
        for k1 in range(K):
            for k2 in range(N):
                iA = dicomSeries[c0].pixel_array
                A[:, :, k1, k2] = iA
                c0 += 1
        dt = dicomSeries.getTemporalResolution()
        ipp = dicomSeries.getImagePositionPatient_np(0)
        sliceVec = dicomSeries.getSliceNormalVector()
        self._meta = {
                    'PixelSpacing': [dicomSeries.getDeltaRow()*mm_to_m, dicomSeries.getDeltaCol()*mm_to_m],
                    'SpacingBetweenSlices': dicomSeries.getDeltaSlice()*mm_to_m,
                    'SliceThickness': dicomSeries.getTag('SliceThickness', convertToType=float, ifNotFound=dicomSeries.getDeltaSlice())*mm_to_m,
                    'SliceLocation0': dicomSeries.getTag('SliceLocation', 0, ifNotFound=0.0, convertToType=float)*mm_to_m,
                    'ImagePositionPatient': [i*mm_to_m for i in ipp], 
                    'ImageOrientationPatient': dicomSeries.getTag('ImageOrientationPatient'), 
                    'PatientPosition': dicomSeries.getTag('PatientPosition'), 
                    'Times': [dt*n*ms_to_s for n in range(N)], # ms to s
                    'Dimensions': A.shape,
                    'SliceVector': sliceVec,
                }
        self._updateMatrix()

    def initFromVTI(self, vtiObj, scaleFactor=1.0):
        dx,dy,dz = vtiObj.GetSpacing()
        oo = vtiObj.GetOrigin()
        # 1st option from meta, then field data then default
        try:
            iop = vtkfilters.getFieldData(vtiObj, 'ImageOrientationPatient')
        except AttributeError:
            iop = [1.0,0.0,0.0,0.0,1.0,0.0]
        try:
            sliceVec = vtkfilters.getFieldData(vtiObj, 'SliceVector')
        except AttributeError:
            sliceVec = [0.0,0.0,1.0]
        self._meta = {'PixelSpacing': [dy*scaleFactor, dx*scaleFactor],
                            'ImagePositionPatient': [i*scaleFactor for i in oo],
                            'ImageOrientationPatient': iop,
                            'SpacingBetweenSlices': dz*scaleFactor,
                            'SliceVector': sliceVec,
                            'Dimensions': vtiObj.GetDimensions(),
                            'SliceThickness': dz*scaleFactor,
                            'SliceLocation0': 0.0}
        self._updateMatrix()

    def initFromDicomSeg(self, dicomSeg, scaleFactor=1.0): # TODO
        pass
        # sliceThick = dicomSeg.SharedFunctionalGroupsSequence[0].PixelMeasuresSequence[0].SliceThickness
        # pixSpace = dicomSeg.SharedFunctionalGroupsSequence[0].PixelMeasuresSequence[0].PixelSpacing
        # ipp = [i.PlanePositionSequence[0].ImagePositionPatient for i in dicomSeg.PerFrameFunctionalGroupsSequence]
        # oo = np.array(ipp[0])
        # normalVector = np.array(ipp[-1]) - oo 
        # normalVector = normalVector / np.linalg.norm(normalVector)
        # oo = dicomSeg.PerFrameFunctionalGroupsSequence[0].PlanePositionSequence[0].ImagePositionPatient
        # iop = dicomSeg.SharedFunctionalGroupsSequence[0].PlaneOrientationSequence[0].ImageOrientationPatient
        # seg_data = dicomSeg.pixel_array
        # seg_data = np.transpose(seg_data, axes=[2,1,0])
        # self._meta = {"ImagePositionPatient": [oo[0]*scaleFactor, oo[1]*scaleFactor, oo[2]*scaleFactor], 
        #         "PixelSpacing": [pixSpace[0]*scaleFactor, pixSpace[1]*scaleFactor],
        #         "SliceThickness": sliceThick*scaleFactor,
        #         "SpacingBetweenSlices": sliceThick*scaleFactor,
        #         "Dimensions": seg_data.shape,
        #         "ImageOrientationPatient": iop, 
        #         "SliceVector": normalVector   
        #         }
        # self._updateMatrix()

    def _updateMatrix(self):
        self._matrix = self.buildImage2PatientCoordinateMatrix()

    def buildImage2PatientCoordinateMatrix(self):
        dr, dc, dz = self.Spacing
        oo = self.ImagePositionPatient
        iop = np.array(self.ImageOrientationPatient)
        vZ = self.SliceVector
        matrix = np.array([
            [iop[0]*dc, iop[3]*dr, vZ[0]*dz, oo[0]], 
            [iop[1]*dc, iop[4]*dr, vZ[1]*dz, oo[1]], 
            [iop[2]*dc, iop[5]*dr, vZ[2]*dz, oo[2]], 
            [0, 0, 0, 1]
        ])
        return matrix

    def getMatrix(self):
        return self._matrix

    def getMetaForVTK(self):
        return {
            'Origin': self.Origin,
            'Spacing': self.Spacing,
            'ImageOrientationPatient': self.ImageOrientationPatient,
            'SliceVector': self.SliceVector,
            'Dimensions': self.Dimensions[:3]
        }

    def getMetaForDICOM(self):
        """
        Returns a dictionary with the meta data for DICOM
        Keys are: 
        'ImagePositionPatient', 'PixelSpacing', 'ImageOrientationPatient', 'SliceVector', 'SliceThickness', 'SliceLocation0', 'SpacingBetweenSlices'
        All in mm
        """
        return {
            'ImagePositionPatient': np.array([i*m_to_mm for i in self.ImagePositionPatient]),
            'PixelSpacing': np.array([i*m_to_mm for i in self.PixelSpacing]),
            'ImageOrientationPatient': self.ImageOrientationPatient,
            'SliceVector': self.SliceVector,
            'SliceThickness': self.SliceThickness*m_to_mm,
            'SliceLocation0': self.SliceLocation0*m_to_mm,
            'SpacingBetweenSlices': self.SpacingBetweenSlices*m_to_mm
        }   

    def imageToPatientCoordinates(self, imageCoords):
        if imageCoords.ndim == 1:
            homogeneous_coords = np.hstack((imageCoords, np.ones((1,))))
            return (self._matrix @ homogeneous_coords.T).T[:3]
        else:
            homogeneous_coords = np.hstack((imageCoords, np.ones((imageCoords.shape[0], 1))))
            return (self._matrix @ homogeneous_coords.T).T[:, :3]

    def patientToImageCoordinates(self, patientCoords):
        homogeneous_coords = np.hstack((patientCoords, np.ones((patientCoords.shape[0], 1))))
        return (np.linalg.inv(self._matrix) @ homogeneous_coords.T).T[:, :3]

    def getVtkMatrix(self):
        vtkMatrix = vtk.vtkMatrix4x4()
        for i in range(4):
            for j in range(4):
                vtkMatrix.SetElement(i, j, self._matrix[i, j])
        return vtkMatrix

    def getVTKTransform(self):
        vtkMatrix = self.getVtkMatrix()
        vtkTransform = vtk.vtkTransform()
        vtkTransform.SetMatrix(vtkMatrix)
        return vtkTransform

    def transformVTKData(self, vtkData):
        transform = self.getVTKTransform()
        transformFilter = vtk.vtkTransformFilter()
        transformFilter.SetInputData(vtkData)
        transformFilter.SetTransform(transform)
        transformFilter.Update()
        return transformFilter.GetOutput()

    def getMeta(self):
        return self._meta

    def setMeta(self, key, value):
        self._meta[key] = value
        self._updateMatrix()

    def updateFromMeta(self, metaDict):
        self._meta.update(metaDict)
        self._updateMatrix()


# ===================================================================================================
# EXPOSED METHODS
# ===================================================================================================

def arrToVTI(arr: np.ndarray, 
            patientMeta: PatientMeta, 
            ds: Optional[dicom.Dataset] = None, 
            TRUE_ORIENTATION: bool = False,
            outputPath: Optional[str] = None) -> Dict[float, vtk.vtkImageData]:
    """Convert array (+meta) to VTI dict (keys=times, values=VTI volumes). 

    Args:
        arr (np.array): Array of pixel data, shape: nR,nC,nSlice,nTime
        patientMeta (PatientMatrix): PatientMatrix object containing meta to be added as Field data
        ds (pydicom dataset [optional]): pydicom dataset to use to add dicom tags as field data
        TRUE_ORIENTATION (bool [False]) : Boolean to force accurate spatial location of image data.
                                NOTE: this uses resampling from VTS data so output VTI will have different dimensions. 
        outputPath (str [None]): Save as go, then rename at end - RAM friendly
    
    Returns:
        vtiDict

    Raises:
        ValueError: If VTK import not available
    """
    dims = arr.shape
    vtkDict = {}
    for k1 in range(dims[-1]):
        A3 = arr[:,:,:,k1]
        ###        
        try:
            thisTime = patientMeta.Times[k1]
        except KeyError:
            thisTime = k1
        if TRUE_ORIENTATION:
            A3 = np.swapaxes(A3, 0, 1)
            vts_data = __arr3ToVTS(A3, patientMeta, ds, thisTime=thisTime)
            newImg = filterResampleToImage(vts_data, np.min(patientMeta.Spacing))
            vtkfilters.delArraysExcept(newImg, [], pointData=False)
            vtkfilters.delArraysExcept(newImg, ['PixelData'])
        else:
            # A3 = np.swapaxes(A3, 0, 1)
            newImg = _arrToImagedata(A3, patientMeta)

        if ds is not None:
            addFieldDataFromDcmDataSet(newImg, ds, extra_tags={"SliceVector": patientMeta.SliceVector,
                                                                "Time": thisTime})
        if outputPath is not None:
            newImg = fIO.writeVTKFile(newImg, os.path.join(outputPath, f"data_{generate_uid()}_{k1:05d}.vti"))
        vtkDict[thisTime] = newImg
    return vtkDict

def _arrToImagedata(A3: np.ndarray, patientMeta: PatientMeta) -> vtk.vtkImageData:
    newImg = _buildVTIImage(patientMeta)
    npArray = np.reshape(A3, np.prod(A3.shape), 'F').astype(np.int16)
    aArray = numpy_support.numpy_to_vtk(npArray, deep=1)
    aArray.SetName('PixelData')
    newImg.GetPointData().SetScalars(aArray)
    return newImg

def _buildVTIImage(patientMeta: PatientMeta=None) -> vtk.vtkImageData:
    if patientMeta is None:
        patientMeta = PatientMeta()
    vti_image = vtk.vtkImageData()
    vti_image.SetSpacing(patientMeta.Spacing[0], patientMeta.Spacing[1], patientMeta.Spacing[2])
    vti_image.SetOrigin(patientMeta.Origin[0], patientMeta.Origin[1], patientMeta.Origin[2])
    vti_image.SetDimensions(patientMeta.Dimensions[0], patientMeta.Dimensions[1], patientMeta.Dimensions[2])
    return vti_image


def __arr3ToVTS(arr: np.ndarray, patientMeta: PatientMeta, ds: Optional[dicom.Dataset] = None, thisTime: Optional[float]=0.0) -> vtk.vtkStructuredGrid:
    dummyPatientMeta = PatientMeta()
    dummyPatientMeta.setMeta('Dimensions', arr.shape)
    ii = _arrToImagedata(arr, patientMeta=dummyPatientMeta) # No info here, let the transform take care of it
    vts_data = patientMeta.transformVTKData(ii)
    if ds is not None:
        addFieldDataFromDcmDataSet(vts_data, ds, extra_tags={"SliceVector": patientMeta.SliceVector,
                                                                "Time": thisTime})
    return vts_data


def arrToVTS(arr: np.ndarray, 
            patientMeta: PatientMeta, 
            ds: Optional[dicom.Dataset] = None, 
            outputPath: Optional[str] = None) -> Dict[float, vtk.vtkStructuredGrid]:
    dims = arr.shape
    vtkDict = {}
    for k1 in range(dims[-1]):
        A3 = arr[:,:,:,k1]
        A3 = np.swapaxes(A3, 0, 1)
        try:
            thisTime = patientMeta.Times[k1]
        except KeyError:
            thisTime = k1
        vts_data = __arr3ToVTS(A3, patientMeta, ds, thisTime)
        if outputPath is not None:
            vts_data = fIO.writeVTKFile(vts_data, os.path.join(outputPath, f"data_{generate_uid()}_{k1:05d}.vts"))
        vtkDict[thisTime] = vts_data
    return vtkDict


def writeArrToVTI(arr: np.ndarray, patientMeta: PatientMeta, filePrefix: str, outputPath: str, ds: Optional[dicom.Dataset] = None, TRUE_ORIENTATION: bool = False) -> str:
    """Will write a VTI file(s) from arr (if np.ndim(arr)=4 write vti files + pvd file)

    Args:
        arr (np.array): Array of pixel data, shape: nR,nC,nSlice,nTime
        patientMeta (PatientMatrix): PatientMatrix object containing meta to be added as Field data
        filePrefix (str): File name prefix (if nTime>1 then named '{fileprefix}_{timeID:05d}.vti)
        outputPath (str): Output path (if nTime > 1 then '{fileprefix}.pvd written to outputPath and sub-directory holds `*.vti` files)
        ds (pydicom dataset [optional]): pydicom dataset to use to add dicom tags as field data

    Raises:
        ValueError: If VTK import not available
    """
    vtkDict = arrToVTI(arr, patientMeta, ds=ds, TRUE_ORIENTATION=TRUE_ORIENTATION)
    return writeVTIDict(vtkDict, outputPath, filePrefix)

def writeVTIDict(vtiDict: Dict[float, vtk.vtkImageData], outputPath: str, filePrefix: str) -> str:
    times = sorted(vtiDict.keys())
    if len(times) > 1:
        return fIO.writeVTK_PVD_Dict(vtiDict, outputPath, filePrefix, 'vti', BUILD_SUBDIR=True)
    else:
        fOut = os.path.join(outputPath, f'{filePrefix}.vti')
        return fIO.writeVTKFile(vtiDict[times[0]], fOut)

def scaleVTI(vti_data: vtk.vtkImageData, factor: float) -> None:
    vti_data.SetOrigin([i*factor for i in vti_data.GetOrigin()])
    vti_data.SetSpacing([i*factor for i in vti_data.GetSpacing()])

def filterResampleToImage(vtsObj: vtk.vtkStructuredGrid, target_spacing: List[float]) -> vtk.vtkStructuredGrid:
    rif = vtk.vtkResampleToImage()
    rif.SetInputDataObject(vtsObj)    
    try:
        target_spacing[0]
    except IndexError:
        target_spacing = [target_spacing, target_spacing, target_spacing]
    bounds = vtsObj.GetBounds()
    dims = [
        int((bounds[1] - bounds[0]) / target_spacing[0]),
        int((bounds[3] - bounds[2]) / target_spacing[1]),
        int((bounds[5] - bounds[4]) / target_spacing[2])
    ]
    rif.SetSamplingDimensions(dims[0],dims[1],dims[2])
    rif.Update()
    return rif.GetOutput()


def readImageStackToVTI(imageFileNames: List[str], patientMeta: PatientMeta=None, arrayName: str = 'PixelData', CONVERT_TO_GREYSCALE: bool = False) -> vtk.vtkImageData:
    append_filter = vtk.vtkImageAppend()
    append_filter.SetAppendAxis(2)  # Combine images along the Z axis
    for file_name in imageFileNames:
        thisImage = fIO.readVTKFile(file_name)
        append_filter.AddInputData(thisImage)
    append_filter.Update()
    combinedImage = append_filter.GetOutput()
    if patientMeta is None:
        patientMeta = PatientMeta()
    combinedImage.SetOrigin(patientMeta.Origin)
    combinedImage.SetSpacing(patientMeta.Spacing)
    a = vtkfilters.getScalarsAsNumpy(combinedImage, RETURN_3D=True)
    if CONVERT_TO_GREYSCALE:
        a = np.mean(a, -1)
    # Manipulation required to bring jpgs to same orientation as vti
    a = np.rot90(a, k=-1) 
    a = np.flipud(a)
    vtkfilters.setArrayFromNumpy(combinedImage, a, arrayName, SET_SCALAR=True, IS_3D=True)
    vtkfilters.delArraysExcept(combinedImage, [arrayName])
    return combinedImage


def _process_phase_time_point(args):
    """Helper function for parallel processing of 4D flow time points
    """
    iTime, magFile, phaseFiles_dicts, phase_factors, phase_offsets, scale_factor, velArrayName, rootDir, fName = args
    iVTS = fIO.readVTKFile(magFile)
    thisVelArray = []
    for k2, iPhase in enumerate(phaseFiles_dicts):
        thisPhase_time = ftk.getClosestFloat(iTime, list(iPhase.keys()))
        thisPhase = fIO.readVTKFile(iPhase[thisPhase_time])
        aName = vtkfilters.getScalarsArrayName(thisPhase)
        thisVelArray.append(vtkfilters.getArrayAsNumpy(thisPhase, aName)*phase_factors[k2] + phase_offsets[k2])
    thisVelArray_ = np.array(thisVelArray).T
    thisVelArray_ *= scale_factor
    vtkfilters.setArrayFromNumpy(iVTS, thisVelArray_, velArrayName, SET_VECTOR=True)
    fOutTemp = fIO.writeVTKFile(iVTS, os.path.join(rootDir, f"{fName}_{generate_uid()}.WORKING.vts"))
    return (iTime, fOutTemp)

def mergePhaseSeries4D(magPVD, phasePVD_list, outputFileName, phase_factors, phase_offsets, 
                        scale_factor=0.001, velArrayName="Vels", DEL_ORIGINAL=True):
    """Take Mag PVD (vts format) and phase PVDs (vts format) and merge into 4D flow dataset

    Args:
        magPVD (str): Path to magnitude PVD file
        phasePVD_list (list): List of paths to phase PVD files
        outputFileName (str): Name of output file
        phase_factors (list): List of factors to multiply phases by. 
        phase_offsets (list): List of offsets to add to phases.
        scale_factor (float): Scale factor to multiply final velocities by. Default is 0.001 (match vel to spatial units)
        velArrayName (str): Name of velocity array to use in output file. Default is "Vels"
        DEL_ORIGINAL (bool, optional): Delete original/intermediate files. Defaults to True.

    Returns:
        str: Name of output file
    """
    rootDir, fName = os.path.split(outputFileName)
    fName = os.path.splitext(fName)[0]
    magFiles = fIO.readPVDFileName(magPVD)
    phaseFiles_dicts = [fIO.readPVDFileName(i) for i in phasePVD_list]
    times = sorted(magFiles.keys())
    outputFilesDict = {}
    # Create process pool and process all time points in parallel
    with Pool() as pool:
        args = [(iTime, magFiles[iTime], phaseFiles_dicts, phase_factors, phase_offsets, 
                scale_factor, velArrayName, rootDir, fName) 
                for iTime in times]
        
        results = pool.map(_process_phase_time_point, args)
    # Collect results into outputFilesDict
    outputFilesDict = dict(results)
    # Generate final output file here - then clean up intermediate files
    pvdFileOut = fIO.writeVTK_PVD_Dict(outputFilesDict, rootDir, fName, "vts", BUILD_SUBDIR=True)

    if DEL_ORIGINAL:
        fIO.deleteFilesByPVD(magPVD)
        for iPhaseFile in phasePVD_list:
            fIO.deleteFilesByPVD(iPhaseFile)
    return pvdFileOut

# ===================================================================================================
# ===================================================================================================
## DICOM-SEG
# ===================================================================================================
def vti_to_dcm_seg(vtiFile, labelMapArrayName, source_dicom_ds_list, dcmSegFileOut=None, algorithm_identification=None):
    imageData = fIO.readVTKFile(vtiFile)
    arr = vtkfilters.getArrayAsNumpy(imageData, labelMapArrayName)
    arr = np.transpose(np.squeeze(arr), axes=[2,0,1])
    return array_to_DcmSeg(arr, source_dicom_ds_list, dcmSegFileOut=dcmSegFileOut, algorithm_identification=algorithm_identification)


def array_to_DcmSeg(arr, source_dicom_ds_list, dcmSegFileOut=None, algorithm_identification=None):
    """Convert a numpy array to DICOM-SEG format.
    
    Args:
        arr: Numpy array of segmentation labels
        source_dicom_ds_list: List of source DICOM datasets
        dcmSegFileOut: Output filename (optional)
        algorithm_identification: Algorithm identification sequence (optional)
        
    Returns:
        DICOM-SEG dataset or filename if dcmSegFileOut is provided
    """

    fullLabelMap = arr.astype(np.ushort)
    sSeg = sorted(set(fullLabelMap.flatten('F')))
    sSeg.remove(0)
    sSegDict = {}
    for k1, segID in enumerate(sSeg):
        sSegDict[k1+1] = f"Segment{k1+1}"
        fullLabelMap[fullLabelMap==segID] = k1+1

    # Describe the algorithm that created the segmentation if not given
    if algorithm_identification is None:
        algorithm_identification = AlgorithmIdentificationSequence(
            name='Spydcmtk',
            version='1.0',
            family=codes.cid7162.ArtificialIntelligence
        )
    segDesc_list = []
    # Describe the segment
    for segID, segName in sSegDict.items():
        description_segment = SegmentDescription(
            segment_number=segID,
            segment_label=segName,
            segmented_property_category=codes.cid7150.Tissue,
            segmented_property_type=codes.cid7154.Kidney,
            algorithm_type=SegmentAlgorithmTypeValues.AUTOMATIC,
            algorithm_identification=algorithm_identification,
            tracking_uid=generate_uid(),
            tracking_id='spydcmtk %s'%(segName)
        )
        segDesc_list.append(description_segment)
    
    try:
        # Create the Segmentation instance
        seg_dataset = Segmentation(
            source_images=source_dicom_ds_list,
            pixel_array=fullLabelMap,
            segmentation_type=SegmentationTypeValues.BINARY,
            segment_descriptions=segDesc_list,
            series_instance_uid=generate_uid(),
            series_number=2,
            sop_instance_uid=generate_uid(),
            instance_number=1,
            manufacturer='Manufacturer',
            manufacturer_model_name='Model',
            software_versions='v1',
            device_serial_number='Device XYZ',
        )
    except Exception as e:
        print(f"Error creating segmentation: {str(e)}")
        print(f"Number of source images: {len(source_dicom_ds_list)}")
        print(f"Label map shape: {fullLabelMap.shape}")
        raise

    if dcmSegFileOut is not None:
        seg_dataset.save_as(dcmSegFileOut) 
        return dcmSegFileOut
    return seg_dataset


def getDcmSeg_meta_depreciated(dcmseg):
    sliceThick = dcmseg.SharedFunctionalGroupsSequence[0].PixelMeasuresSequence[0].SliceThickness
    pixSpace = dcmseg.SharedFunctionalGroupsSequence[0].PixelMeasuresSequence[0].PixelSpacing
    ffgs = dcmseg.PerFrameFunctionalGroupsSequence
    ipp = [i.PlanePositionSequence[0].ImagePositionPatient for i in dcmseg.PerFrameFunctionalGroupsSequence]
    oo = np.array(ipp[0])
    normalVector = np.array(ipp[-1]) - oo 
    normalVector = normalVector / np.linalg.norm(normalVector)
    oo = dcmseg.PerFrameFunctionalGroupsSequence[0].PlanePositionSequence[0].ImagePositionPatient
    iop = dcmseg.SharedFunctionalGroupsSequence[0].PlaneOrientationSequence[0].ImageOrientationPatient
    seg_data = dcmseg.pixel_array
    seg_data = np.transpose(seg_data, axes=[2,1,0])
    return {"Origin": [oo[0]*0.001, oo[1]*0.001, oo[2]*0.001], 
            "Spacing": [pixSpace[0]*0.001, pixSpace[1]*0.001, sliceThick*0.001],
            "Dimensions": seg_data.shape,
            "ImageOrientationPatient": iop, 
            "SliceVector": normalVector   
            }


# def dicom_seg_to_vtk_depreciated(dicom_seg_path, vtk_output_path, TRUE_ORIENTATION=False):
#     ds = dicom.dcmread(dicom_seg_path)
#     patMeta = PatientMeta()
#     patMeta.initFromDicomSeg(ds)
#     seg_data = ds.pixel_array
#     seg_data = np.transpose(seg_data, axes=[2,1,0])
#     image_data = vtk.vtkImageData()
#     image_data.SetOrigin(patMeta.Origin)
#     image_data.SetDimensions(patMeta.Dimensions)
#     image_data.SetSpacing(patMeta.Spacing)
#     vtk_array = numpy_support.numpy_to_vtk(num_array=seg_data.flatten('F'), deep=True, array_type=vtk.VTK_UNSIGNED_CHAR)
#     image_data.GetPointData().SetScalars(vtk_array)
#     if TRUE_ORIENTATION:
#         fIO.writeVTS(_vti2vts(image_data, patMeta), vtk_output_path)
#     else:
#         fIO.writeVTI(image_data, vtk_output_path)
#     return vtk_output_path


class NoVtkError(Exception):
    ''' NoVtkError
            If VTK import fails '''
    def __init__(self):
        pass
    def __str__(self):
        return 'NoVtkError: VTK not found. Run: "pip install vtk"'


# ===================================================================================================
# ===================================================================================================



# ===================================================================================================
# ===================================================================================================
## HELPFUL FILTERS
# ===================================================================================================

def addFieldDataFromDcmDataSet(vtkObj, ds, extra_tags={}):
    tagsDict = dcmTools.getDicomTagsDict()
    for iTag in tagsDict.keys():
        try:
            val = ds[iTag].value
            if type(val) in [dicom.multival.MultiValue, dicom.valuerep.DSfloat, dicom.valuerep.IS]:
                try:
                    tagArray = numpy_support.numpy_to_vtk(np.array(val))
                except TypeError: # multivalue - but prob strings
                    tagArray = vtk.vtkStringArray()
                    tagArray.SetNumberOfValues(len(val))
                    for k1 in range(len(val)):
                        tagArray.SetValue(k1, str(val[k1]))
            else:
                tagArray = vtk.vtkStringArray()
                tagArray.SetNumberOfValues(1)
                tagArray.SetValue(0, str(val))
            tagArray.SetName(iTag)
            vtkObj.GetFieldData().AddArray(tagArray)
        except KeyError:
            continue # tag not found
    for iTag in extra_tags:
        val = extra_tags[iTag]
        tagArray = numpy_support.numpy_to_vtk(np.array(val))
        tagArray.SetName(iTag)
        vtkObj.GetFieldData().AddArray(tagArray)



# -*- coding: utf-8 -*-

"""Classes for working with Dicom studies
"""
import copy
import os
import pydicom as dicom
from pydicom.uid import generate_uid
from datetime import datetime
from tqdm import tqdm
import json
from multiprocessing import Pool
import numpy as np
import shutil
import matplotlib.pyplot as plt
# Local imports 
import spydcmtk.dcmTools as dcmTools
import spydcmtk.dcmVTKTK as dcmVTKTK
from spydcmtk.spydcm_config import SpydcmTK_config




## =====================================================================================================================
##        CLASSES
## =====================================================================================================================
class DicomSeries(list):
    """Extends a list of ds (pydicom dataset) objects.
    """
    def __init__(self, dsList=None, OVERVIEW=False, HIDE_PROGRESSBAR=False, FORCE_READ=False, SAFE_NAME_MODE=False):
        """Initialise DicomSeries with list of pydicom datasets.

        Args:
            dsList (list, optional): list of pydicom dataset. Defaults to None.
            OVERVIEW (bool, optional): Set True to not read pixel data, for efficiency in applications of meta 
                                        data only interest. 
                                        Pass to pydicom.filereader.dcmread parameter:stop_before_pixels
                                        Defaults to False.
            HIDE_PROGRESSBAR (bool, optional): Set True to hide tqdm progressbar. Defaults to False.
            FORCE_READ (bool, optional): Force read dicom files even if they do not conform to DICOM standard. 
                                            Pass to pydicom.filereader.dcmread parameter:force
                                            Defaults to False.
            SAFE_NAME_MODE (bool, optional): Set writing mode to use UIDs for file naming. Defaults to False.
        """
        if dsList is None:
            dsList = []
        self.OVERVIEW = OVERVIEW
        self.HIDE_PROGRESSBAR = HIDE_PROGRESSBAR
        self.FORCE_READ = FORCE_READ
        self.SAFE_NAME_MODE = SAFE_NAME_MODE
        self.NOT_FULLY_LOADED = False
        list.__init__(self, dsList)

    def __str__(self):
        """Return output of getSeriesOverview method
        """
        return ' '.join([str(i) for i in self.getSeriesOverview()[1]])


    def __getitem__(self, item):
        if isinstance(item, slice):
            result = super().__getitem__(item)
            return DicomSeries(result)
        else:
            return super().__getitem__(item)

    @classmethod
    def _setFromDictionary(cls, dicomDict, OVERVIEW=False, HIDE_PROGRESSBAR=False, FORCE_READ=False):
        dStudyList = ListOfDicomStudies.setFromDcmDict(dicomDict, OVERVIEW=OVERVIEW, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)
        if len(dStudyList) > 1:
            raise ValueError('More than one study found - use ListOfDicomStudies class')
        try:
            dStudy = dStudyList[0]
        except IndexError:
            raise ValueError('No DICOMS found - please check your inputs')
        if len(dStudy) > 1:
            raise ValueError('More than one series found - use DicomStudy class')
        return cls(dStudy[0], OVERVIEW=OVERVIEW, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)

    @classmethod
    def setFromDirectory(cls, dirName, OVERVIEW=False, HIDE_PROGRESSBAR=False, FORCE_READ=False, ONE_FILE_PER_DIR=False):
        """Initialise object from directory of dicom files.

        Args:
            dirName (str): Path to directory of dicom files
            OVERVIEW (bool, optional): Set attribute OVERVIEW. Defaults to False.
            HIDE_PROGRESSBAR (bool, optional): Set attribute HIDE_PROGRESSBAR. Defaults to False.
            FORCE_READ (bool, optional): Set attribute FORCE_READ. Defaults to False.
            ONE_FILE_PER_DIR (bool, optional): Read only one file per directory for fast applications
                                                with meta interest only. Defaults to False.

        Raises:
            ValueError: If dicom files belonging to more than one study are found.
            ValueError: If dicom files belonging to more than one series are found.

        Returns:
            DicomSeries: An instance of DicomSeries class.
        """
        dicomDict = dcmTools.organiseDicomHeirarchyByUIDs(dirName, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ, ONE_FILE_PER_DIR=ONE_FILE_PER_DIR, OVERVIEW=OVERVIEW)
        obj = DicomSeries._setFromDictionary(dicomDict, OVERVIEW=OVERVIEW, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)
        obj.NOT_FULLY_LOADED = ONE_FILE_PER_DIR
        return obj

    @classmethod
    def setFromFileList(cls, fileList, OVERVIEW=False, HIDE_PROGRESSBAR=False, FORCE_READ=False):
        """Initialise object from list of files

        Args:
            fileList (list): List of full file names
            OVERVIEW (bool, optional): Set attribute OVERVIEW. Defaults to False.
            HIDE_PROGRESSBAR (bool, optional): Set attribute HIDE_PROGRESSBAR. Defaults to False.
            FORCE_READ (bool, optional): Set attribute FORCE_READ. Defaults to False.

        Raises:
            ValueError: If dicom files belonging to more than one study are found.
            ValueError: If dicom files belonging to more than one series are found.

        Returns:
            DicomSeries: An instance of DicomSeries class.
        """
        dicomDict = {}
        for iFile in fileList:
            dcmTools.readDicomFile_intoDict(iFile, dicomDict, FORCE_READ=FORCE_READ, OVERVIEW=OVERVIEW)
        return DicomSeries._setFromDictionary(dicomDict, OVERVIEW=OVERVIEW, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)

    def getRootDir(self):
        return os.path.split(self[0].filename)[0]

    def filterByTag(self, tagName, tagValue):
        return DicomSeries([i for i in self if i.getTag(tagName) == tagValue], 
                           OVERVIEW=self.OVERVIEW, 
                           HIDE_PROGRESSBAR=self.HIDE_PROGRESSBAR, 
                           FORCE_READ=self.FORCE_READ)

    def getDicomFullFileName(self, dsID=0):
        return self[dsID].filename

    def sortByInstanceNumber(self):
        self.sort(key=dcmTools.instanceNumberSortKey)

    def sortBySlice_InstanceNumber(self):
        self.sort(key=dcmTools.sliceLoc_InstanceNumberSortKey)

    def getTag(self, tag, dsID=0, ifNotFound='Unknown', convertToType=None):
        try:
            tt = self.getTagObj(tag, dsID)
            if convertToType is not None:
                return convertToType(tt.value)
            return tt.value
        except KeyError:
            return ifNotFound

    def setTags_all(self, tag, value):
        for ds in self:
            ds[tag].value = value

    def getTagObj(self, tag, dsID=0):
        try:
            if tag[:2] == '0x':
                tag = int(tag, 16)
        except TypeError:
            pass # Not string
        return self[dsID][tag]

    def getTagValuesList(self, tagList, RETURN_STR):
        """
        Build table with file name and value from each tag in list.
        Return list of lists
        """
        valueList = []
        for dsID, i in enumerate(self):
            subList = ['"%s"'%(str(i.filename))] + [str(self.getTag(t, dsID=dsID)) for t in tagList]
            valueList.append(subList)
        if RETURN_STR:
            return dcmTools._tagValuesListToString(tagList, valueList)
        return valueList

    def getTagListAndNames(self, tagList, dsID=0):
        names, vals = [], []
        for i in tagList:
            try:
                dataEle = self.getTagObj(i, dsID=dsID)
                iname = str(dataEle.keyword)
                if len(iname) == 0:
                    iname = str(dataEle.name)
                    iname = iname.replace('[','').replace(']','').replace(' ','')
                names.append(iname)
                vals.append(str(dataEle.value))
            except KeyError:
                names.append(str(i))
                vals.append('Unknown')
        return names, vals

    def tagsToJson(self, jsonFileOut):
        dOut = {}
        for ds in self:
            dd = ds.to_json_dict()
            dd.pop('7FE00010') # Remove actual PixelData
            dOut[ds.InstanceNumber] = dd
        dcmTools.writeDictionaryToJSON(jsonFileOut, dOut)
        return jsonFileOut


    def _loadToMemory(self):
        if self.NOT_FULLY_LOADED:
            rootDir = self.getRootDir()
            self.clear()
            self.extend([dicom.dcmread(os.path.join(rootDir, i), force=self.FORCE_READ) for i in tqdm(os.listdir(rootDir), disable=self.HIDE_PROGRESSBAR)])
            self.NOT_FULLY_LOADED = False


    def getSeriesOverview(self, tagList=SpydcmTK_config.SERIES_OVERVIEW_TAG_LIST):
        names, vals = self.getTagListAndNames(tagList)
        names.append('ImagesInSeries')
        if len(self) == 1: # ONLY READ ONE FILE PER DIR
            vals.append(dcmTools.countFilesInDir(self.getRootDir()))
        else:
            vals.append(len(self))
        return names, vals

    def getSeriesTimeAsDatetime(self):
        dos = self.getTag('SeriesDate', ifNotFound="19000101")
        tos = self.getTag('SeriesTime', ifNotFound="000000")
        try:
            return datetime.strptime(f"{dos} {tos}", "%Y%m%d %H%M%S.%f")
        except ValueError:
            return datetime.strptime(f"{dos} {tos}", "%Y%m%d %H%M%S")

    def getImagePositionPatient_np(self, dsID):
        ipp = self.getTag('ImagePositionPatient', dsID=dsID, ifNotFound=[0.0,0.0,0.0])
        return np.array(ipp)

    def getImageOrientationPatient_np(self, dsID=0):
        iop = self.getTag('ImageOrientationPatient', dsID=dsID)
        return np.array(iop)

    def yieldDataset(self):
        for ds in self:
            if self.OVERVIEW:
                yield  dicom.dcmread(ds.filename, stop_before_pixels=False, force=True)
            else:
                yield ds

    def isCompressed(self):
        return dcmTools._isCompressed(self[0])

    def getSeriesNumber(self):
        return int(self.getTag('SeriesNumber', ifNotFound=0))

    def getSeriesOutDirName(self, SE_RENAME={}):
        thisSeNum = self.getTag('SeriesNumber', ifNotFound='#')
        suffix = ''
        if (thisSeNum=="#") or self.SAFE_NAME_MODE:
            suffix += '_'+self.getTag("SeriesInstanceUID")
        for iTag in SpydcmTK_config.SERIES_NAMING_TAG_LIST:
            iVal = self.getTag(iTag, ifNotFound='')
            if len(iVal) > 0:
                suffix += '_'+iVal
        if thisSeNum in SE_RENAME.keys():
            return SE_RENAME[thisSeNum]
        return dcmTools.cleanString(f"SE{thisSeNum}{suffix}")

    # ----------------------------------------------------------------------------------------------------
    def removeTimes(self, timesIDs_ToRemove):
        K = int(self.getNumberOfSlicesPerVolume())
        self.sortBySlice_InstanceNumber()
        N = self.getNumberOfTimeSteps()
        c0 = 0
        dsToRm = []
        for k1 in range(K):
            for k2 in range(N):
                if k2 in timesIDs_ToRemove:
                    dsToRm.append(self[c0])
                c0 += 1
        self._rm_via_DS_List(dsToRm)


    def removeInstances(self, instanceIDs_to_remove):
        dsToRm = []
        for k1 in range(len(self)):
            if self.getTag("InstanceNumber", k1, convertToType=int) in instanceIDs_to_remove:
                dsToRm.append(self[k1])
        self._rm_via_DS_List(dsToRm)


    def _rm_via_DS_List(self, dsList):
        for iDS in dsList:
            self.remove(iDS)


    def deleteTag(self, tag, dsID):
        tagObj = self.getTagObj(tag, dsID)
        del tagObj

    # ----------------------------------------------------------------------------------------------------
    def resetUIDs(self, studyUID):
        """Reset SOPInstanceUID, SeriesInstanceUID and StudyInstaceUID (must pass last)

        Args:
            studyUID (str): UID - can use str(generate_uid())
        """
        seriesUID = str(generate_uid())
        for k1 in range(len(self)):
            self[k1].SOPInstanceUID = str(generate_uid())
            self[k1].SeriesInstanceUID = seriesUID
            self[k1].StudyInstanceUID = studyUID

    def anonymise(self, anonName, anonPatientID, anon_birthdate=True, remove_private_tags=True):
        """Anonymise series inplace

        Args:
            anonName (str): New anonymise name (can be empty str)
            anonPatientID (str): New anon patient ID, if None then no action taken.
            anon_birthdate (bool, optional): To remove birthdate from tags. Defaults to True.
            remove_private_tags (bool, optional): To remove private tags from dataset. Defaults to True. 
                In some cases this will need to be set False to keep tags necessary for post processing (e.g. some DTI tags). 
                It is up to the user to confirm complete anonymisation in these cases. 
        """
        def PN_callback(ds, data_element):
            """Called from the dataset "walk" recursive function for all data elements."""
            if data_element.VR == "PN":
                data_element.value = 'anonymous'
            if "Institution" in data_element.name:
                data_element.value = 'anonymous'
            if data_element.name == "Patient's Name":
                data_element.value = anonName
            if (data_element.name == "Patient ID") and (anonPatientID is not None):
                data_element.value = anonPatientID

        # Remove patient name and any other person names using a callback to walk over data elements for each ds
        for dataset in self:
            try:
                dataset.walk(PN_callback)
            except TypeError: 
                pass
            # Other Tags
            for name in ['OtherPatientIDs', 'OtherPatientIDsSequence', 'PatientAddress']:
                if name in dataset:
                    delattr(dataset, name)
            if anon_birthdate:
                for name in ['PatientBirthDate']:
                    if name in dataset:
                        dataset.data_element(name).value = ''
            if remove_private_tags:
                try:
                    dataset.remove_private_tags()
                except TypeError:
                    pass

    def writeToOrganisedFileStructure(self, studyOutputDir, SAFE_NAMING_CHECK=True, seriesOutDirName=None):
        """ Recurse down directory tree - grab dicoms and move to new
            hierarchical folder structure rooted at 'studyOutputDir'
        """
        self._loadToMemory()
        if SAFE_NAMING_CHECK:
            self.checkIfShouldUse_SAFE_NAMING()
        ADD_TRANSFERSYNTAX = False
        if seriesOutDirName is None:
            seriesOutDirName = self.getSeriesOutDirName()
        seriesOutputDir = os.path.join(studyOutputDir, seriesOutDirName)
        seriesOutputDirTemp = seriesOutputDir+".WORKING"
        if os.path.isdir(seriesOutputDir):
            os.rename(seriesOutputDir, seriesOutputDirTemp)
        if self.FORCE_READ:
            ADD_TRANSFERSYNTAX = True # need to add if required force read. 
        TO_DECOMPRESS = self.isCompressed()
        for ds in tqdm(self.yieldDataset(), total=len(self), disable=self.HIDE_PROGRESSBAR):
            if TO_DECOMPRESS:
                try:
                    ds.decompress()
                except AttributeError:
                    print(f'Error with file in {seriesOutputDirTemp}')
                    continue
            if ADD_TRANSFERSYNTAX:
                ds.file_meta.TransferSyntaxUID = '1.2.840.10008.1.2.1'
            destFile = dcmTools.writeOut_ds(ds, seriesOutputDirTemp, SAFE_NAMING=self.SAFE_NAME_MODE)
        # ON COMPLETION RENAME OUTPUTDIR
        os.rename(seriesOutputDirTemp, seriesOutputDir)
        return seriesOutputDir

    def _generateFileName(self, tagsToUse, extn):
        if type(tagsToUse) == str:
            fileName = tagsToUse
        else:
            fileName = '_'.join([str(self.getTag(i)) for i in tagsToUse])
        fileName = dcmTools.cleanString(fileName)
        if (len(extn) > 0) and (not extn.startswith('.')):
            extn = '.'+extn
        return fileName+extn

    def writeToNII(self, outputPath, outputNamingTags=('PatientName', 'SeriesNumber', 'SeriesDescription')):
        """Will write DicomSeries as nii.gz file (uses dcm2niix found on local system)

        Args:
            outputPath (str): Directory where to save output (if full pathname, with extn, is given then this is used)
            outputNamingTags (tuple, optional): Tags to use to name output. If outputPath contains extension then not used. Defaults to ('PatientName', 'SeriesNumber', 'SeriesDescription').

        Returns:
            str: Name of files saved
        """
        if outputPath.endswith('nii.gz'):
            outputPath, fileName = os.path.split(outputPath)
        else:
            fileName = self._generateFileName(outputNamingTags, '.nii.gz')
        return dcmTools.writeDirectoryToNII(self.getRootDir(), outputPath, fileName=fileName)

    def writeToVTI(self, outputPath, outputNamingTags=('PatientName', 'SeriesNumber', 'SeriesDescription'), TRUE_ORIENTATION=False):
        """Write DicomSeries as VTK ImageData (`*.vti`)

        Args:
            outputPath (str): Directory to save in or full filename to save
            outputNamingTags (tuple, optional): Tags to use for naming, only used if full outputpath not given. Defaults to ('PatientName', 'SeriesNumber', 'SeriesDescription').
            TRUE_ORIENTATION (bool, optional): To apply matrix to VTI data. Defaults to False.

        Returns:
            str: Name of file saved
        """
        if outputPath.endswith('vti') or outputPath.endswith('pvd'):
            outputPath, fileName = os.path.split(outputPath)
            fileName, _ = os.path.splitext(fileName)
        else:
            fileName = self._generateFileName(outputNamingTags, '')
        vtiDict = self.buildVTIDict(TRUE_ORIENTATION=TRUE_ORIENTATION)
        return dcmVTKTK.writeVTIDict(vtiDict, outputPath, fileName)

    def writeToVTS(self, outputPath, outputNamingTags=('PatientName', 'SeriesNumber', 'SeriesDescription')):
        """Write DicomSeries as VTK StructuredImageData (`*.vts`)

        Args:
            outputPath (str): Directory to save in or full filename to save.
            outputNamingTags (tuple, optional): Tags to use for naming, only used if full outputpath not given. Defaults to ('PatientName', 'SeriesNumber', 'SeriesDescription').

        Returns:
            str: Name of file saved
        """
        if outputPath.endswith('vts') or outputPath.endswith('pvd'):
            outputPath, fileName = os.path.split(outputPath)
            fileName, _ = os.path.splitext(fileName)
        else:
            fileName = self._generateFileName(outputNamingTags, '')
        vtsDict = self.buildVTSDict(outputPath)
        return dcmVTKTK.fIO.writeVTK_PVD_Dict(vtsDict, outputPath, filePrefix=fileName, fileExtn='vts', BUILD_SUBDIR=True)

    def buildVTSDict(self, outputPath=None):
        A, patientMeta = self.getPixelDataAsNumpy()
        return dcmVTKTK.arrToVTS(A, patientMeta, self[0], outputPath)

    def buildVTIDict(self, TRUE_ORIENTATION=False):
        A, patientMeta = self.getPixelDataAsNumpy()
        return dcmVTKTK.arrToVTI(A, patientMeta, self[0], TRUE_ORIENTATION=TRUE_ORIENTATION)

    @property
    def sliceLocations(self):
        return sorted([float(self.getTag('SliceLocation', i, ifNotFound=0.0)) for i in range(len(self))])

    def getNumberOfSlicesPerVolume(self):
        sliceLoc = self.sliceLocations
        return len(set(sliceLoc))

    def getNumberOfTimeSteps(self):
        sliceLoc = self.sliceLocations
        sliceLocS = set(sliceLoc)
        return sliceLoc.count(sliceLocS.pop())

    def getSliceNormalVector(self):
        self.sortBySlice_InstanceNumber()
        try:
            sliceVec = self.getImagePositionPatient_np(-1) - self.getImagePositionPatient_np(0)
        except:
            sliceVec = [0.0,0.0,0.0] # If no slice locations then assume no normal vector (one image)
        if np.linalg.norm(sliceVec) < 1e-9:
            iop = self.getTag('ImageOrientationPatient', ifNotFound=[1.0,0.0,0.0,0.0,1.0,0.0])
            sliceVec = np.cross(iop[:3], iop[3:6])
        else:
            sliceVec = sliceVec / np.linalg.norm(sliceVec)
        return sliceVec

    def doesSliceLocationNorm_Match_IOPNormal(self):
        iop = self.getTag('ImageOrientationPatient')
        iopN = np.cross(iop[:3], iop[3:6])
        sliceLocN = self.getSliceNormalVector()
        tf = [i*j>=0.0 for i,j in zip(iopN, sliceLocN)]
        return all(tf)

    def is3D(self):
        return self.getNumberOfSlicesPerVolume() > 1

    def is4D(self):
        return self.getNumberOfTimeSteps() > 1

    def getPixelDataAsNumpy(self):
        """Get pixel data as numpy array organised by slice and time(if present).
            Also return a PatientMatrix object containing meta
            NOTE: Some data is repeated to serve VTK and DICOM conventions

        Returns:
            tuple: 
                numpy array - shape [nRow, nCol, nSlice, nTime], 
                dictionary - keys: Spacing, Origin, ImageOrientationPatient, PatientPosition, Dimensions, Times
        """
        self._loadToMemory()
        I,J,K = int(self.getTag('Rows')), int(self.getTag('Columns')), int(self.getNumberOfSlicesPerVolume())
        self.sortBySlice_InstanceNumber() # This takes care of order - slices grouped, then time for each slice 
        N = self.getNumberOfTimeSteps()
        A = np.zeros((I, J, K, N))
        if (K*N) != len(self):
            print(f"DEBUG: Getting numpy array shape: [{I}, {J}, {K}, {N}] (K*N={K*N} == {len(self)})")
            raise FileNotFoundError(f"Missing some DICOM files.")
        c0 = 0
        for k1 in range(K):
            for k2 in range(N):
                iA = self[c0].pixel_array
                A[:, :, k1, k2] = iA
                c0 += 1
        patientMeta = dcmVTKTK.PatientMeta()
        patientMeta.initFromDicomSeries(self)
        return A, patientMeta

    def getScanDuration_secs(self):
        try:
            return self.getTag(0x0019105a, ifNotFound=0.0) / 1000000.0
        except AttributeError:
            return 0.0

    def _getPixelSpacing(self):
        return [float(i) for i in self.getTag('PixelSpacing', ifNotFound=[0.0,0.0])]

    def getDeltaRow(self):
        return self._getPixelSpacing()[0]

    def getDeltaCol(self):
        return self._getPixelSpacing()[1]

    def getDeltaSlice(self):
        self.sortByInstanceNumber()
        p0 = self.getImagePositionPatient_np(0)
        sliceLoc = [distBetweenTwoPts(p0, self.getImagePositionPatient_np(i)) for i in range(len(self))]
        sliceLocS = sorted(list(set(sliceLoc)))
        if len(sliceLocS) == 1: # May be CINE at same location
            dZ = self.getTag('SpacingBetweenSlices', ifNotFound=None)
            if dZ is None:
                dZ = self.getTag('SliceThickness')
            return float(dZ)
        return np.mean(np.diff(sliceLocS))

    def getTemporalResolution(self):
        try:
            return float(self.getTag('NominalInterval', ifNotFound=0.0)/self.getTag('CardiacNumberOfImages', ifNotFound=1))
        except ZeroDivisionError:
            return 0       

    def getTemporalResolution_TR_VPS(self):
        return float(self.getTag('RepetitionTime', ifNotFound=0.0)*self.getTag(0x00431007, ifNotFound=1.0))

    def getManufacturer(self):
        return self.getTag(0x00080070, ifNotFound='Unknown')

    def IS_GE(self):
        return self.getManufacturer().lower().startswith('ge')

    def IS_SIEMENS(self):
        return self.getManufacturer().lower().startswith('siemens')

    def IS_PHILIPS(self):
        return self.getManufacturer().lower().startswith('philips')

    def IS_BRUKER(self):
        return self.getManufacturer().lower().startswith('bruker')

    def getPulseSequenceName(self):
        if self.IS_GE():
            return self.getTag(0x0019109c)
        else:
            return self.getTag(0x00180024)

    def getVENC(self):
        """Get VENC value in mm/s
        """
        if self.IS_GE():
            return float(self.getTag(0x001910cc))
        elif self.IS_SIEMENS():
            vencStr = self.getTag(0x00511014)
            return float(vencStr.split("_")[0].replace("v", "")) * 10.0
        else: # TODO - add other vendors
            return None

    def getInternalPulseSequenceName(self):
        return self.getTag(0x0019109e)

    def getSeriesDescription(self):
        return self.getTag('SeriesDescription')

    def getSeriesInfoDict(self, EXTRA_TAGS=[]):
        # Default (standard tags):
        outDict = {'ScanDuration':self.getScanDuration_secs(),
            'nTime':self.getTag('CardiacNumberOfImages'),
            'nRow':self.getTag('Rows'),
            'nCols':self.getTag('Columns'),
            'dRow':self.getDeltaRow(),
            'dCol':self.getDeltaCol(),
            'dSlice':self.getTag('SliceThickness'),
            'dTime': self.getTemporalResolution(),
            'SpacingBetweenSlices':self.getTag('SpacingBetweenSlices'),
            'FlipAngle':self.getTag('FlipAngle'),
            'HeartRate':self.getTag('HeartRate'),
            'EchoTime':self.getTag('EchoTime'),
            'RepetitionTime':self.getTag('RepetitionTime'),
            'PulseSequenceName':self.getPulseSequenceName(),
            'MagneticFieldStrength': self.getTag('MagneticFieldStrength'),
            'InternalPulseSequenceName':self.getInternalPulseSequenceName(),
            'ReconstructionDiameter': self.getTag(0x00181100),
            'AcquisitionMatrix': str(self.getTag(0x00181310)),
            'Manufacturer': self.getTag("Manufacturer"),
            'ManufacturerModelName': self.getTag("ManufacturerModelName"),
            'SoftwareVersions': str(self.getTag(0x00181020)),}
        for extraTag in EXTRA_TAGS:
            outDict[extraTag] = self.getTag(extraTag)
        outDict['nSlice'] = len(self)
        try:
            outDict['AcquiredResolution'] = float(outDict['ReconstructionDiameter']) / float(max(self.getTag(0x00181310)))
        except ValueError:
            outDict['AcquiredResolution'] = f"{outDict['dRow']},{outDict['dCol']}"
        try:
            outDict['AcquiredTemporalResolution'] = self.getTemporalResolution_TR_VPS()
        except ValueError:
            outDict['AcquiredTemporalResolution'] = 0.0
        for i in outDict.keys():
            try:
                if ',' in outDict[i]:
                    outDict[i] = f'"{outDict[i]}"'
            except TypeError:
                pass
        return outDict

    def checkIfShouldUse_SAFE_NAMING(self, se_instance_set=None):
        """Checks is should name dicoms based on UIDs or if a SE-number, Instace-number naming convention is possible.
        Basically checks if there would be any overlaps. 

        Args:
            se_instance_set (set, optional): A set of instance file names (without extension). Defaults to None.
        """
        if self.SAFE_NAME_MODE:
            return
        if se_instance_set is None:
            se_instance_set = set()
        for k1 in range(len(self)):
            se = self.getTag('SeriesNumber', dsID=k1, ifNotFound='unknown')
            instance = self.getTag('InstanceNumber', dsID=k1, ifNotFound='unknown')
            se_instance_str = f"{se}_{instance}"
            if se_instance_str in se_instance_set:
                self.SAFE_NAME_MODE = True
                break
            se_instance_set.add(se_instance_str)

    def getStudyOutputDir(self, rootDir='', studyPrefix=''):
        """Generate a study output directory name based on config setting STUDY_NAMINF_TAG_LIST

        Args:
            rootDir (str, optional): The root directory to build study directory in. Defaults to ''.
            studyPrefix (str, optional): Optional prefix to add to naming. Defaults to ''.

        Returns:
            str: Path to directory at Study level where dicoms will be written
        """
        suffix = ''
        if self.SAFE_NAME_MODE:
            suffix += f"{self.getTag('StudyInstanceUID')}"
        for iTag in SpydcmTK_config.STUDY_NAMING_TAG_LIST:
            iVal = self.getTag(iTag, ifNotFound='', convertToType=str)
            if len(iVal) > 0:
                suffix += '_'+iVal
        return os.path.join(rootDir, dcmTools.cleanString(studyPrefix+suffix))


    def overviewImage(self, outputFileName=None, RETURN_FIG=False):
        """
        Build an overview image of the series.

        Args:
            outputFileName (str, optional): The name of the file to save the overview image to. Defaults to None.
            RETURN_FIG (bool, optional): Whether to return the figure. Defaults to False.

        Returns:
            if RETURN_FIG is True, returns the figure.
            if outputFileName is not None, saves the figure to the file and returns the file name.
            otherwise, displays the figure.
        """
        A = self.getPixelDataAsNumpy()[0]
        i,j,k,n = A.shape
        if k > 1:
            iSlice = k//2
            Image_list = [A[:,:,iSlice,0], np.rot90(A[i//2,:,:,0],3), np.rot90(A[:,j//2,:,0],3)]
        else:
            Image_list = [A[:,:,0,0]]
        fig, axs = plt.subplots(1, len(Image_list))
        if len(Image_list) == 1:
            axs = [axs]  # Convert single Axes to list for consistent indexing
        for i, iA in enumerate(Image_list):
            axs[i].imshow(iA, cmap='gray')
            axs[i].axis('off')
        fig.tight_layout()
        if RETURN_FIG:
            return fig
        if outputFileName is not None:
            plt.savefig(outputFileName)
            plt.close()
            return outputFileName
        else:
            plt.show()
        


# ================================================================================================================
# DicomStudy
# ================================================================================================================
class DicomStudy(list):
    """
    Extends list of DicomSeries objects (creating list of list of pydicom.dataset)
    """
    def __init__(self, dSeriesList, OVERVIEW=False, HIDE_PROGRESSBAR=False):
        """
        Set OVERVIEW = False to read pixel data as well (at a cost)
        """
        self.OVERVIEW = OVERVIEW
        self.HIDE_PROGRESSBAR = HIDE_PROGRESSBAR # FIXME - if set need ot pass down to series also - for the writing
        list.__init__(self, dSeriesList)

    @classmethod
    def setFromDictionary(cls, dicomDict, OVERVIEW=False, HIDE_PROGRESSBAR=False, FORCE_READ=False):
        dStudyList = ListOfDicomStudies.setFromDcmDict(dicomDict, OVERVIEW=OVERVIEW, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)
        if len(dStudyList) > 1:
            raise ValueError('More than one study found - use ListOfDicomStudies class')
        return cls(dStudyList[0], OVERVIEW=OVERVIEW, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR)

    @classmethod
    def setFromDirectory(cls, dirName, OVERVIEW=False, HIDE_PROGRESSBAR=False, FORCE_READ=False, ONE_FILE_PER_DIR=False):
        dicomDict = dcmTools.organiseDicomHeirarchyByUIDs(dirName, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ, ONE_FILE_PER_DIR=ONE_FILE_PER_DIR, OVERVIEW=OVERVIEW)
        obj = DicomStudy.setFromDictionary(dicomDict, OVERVIEW=OVERVIEW, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)
        for iSeries in obj:
            iSeries.NOT_FULLY_LOADED = ONE_FILE_PER_DIR
        return obj


    def __str__(self):
        return '%s: %s: %d series, %d dicoms'%(self.getTag('PatientName'),
                                                self.getTag('SeriesDescription'),
                                                len(self),
                                                self.getNumberOfDicoms())


    def setSafeNameMode(self):
        for iSeries in self:
            iSeries.SAFE_NAME_MODE = True

    def resetUIDs(self, studyUID=None):
        if studyUID is None:
            studyUID = str(generate_uid())
        for i in self:
            i.resetUIDs(studyUID)

    def filterByTag(self, tagName, tagValue):
        return DicomStudy([i for i in self if i.getTag(tagName) == tagValue], 
                           OVERVIEW=self.OVERVIEW, 
                           HIDE_PROGRESSBAR=self.HIDE_PROGRESSBAR)

    def anonymise(self, anonName, anonPatientID, anonBirthdate=True, removePrivateTags=True):
        for i in self:
            i.anonymise(anonName, anonPatientID, anonBirthdate, removePrivateTags)

    def getTag(self, tag, seriesID=0, instanceID=0, ifNotFound='Unknown', convertToType=None):
        return self[seriesID].getTag(tag, dsID=instanceID, ifNotFound=ifNotFound, convertToType=convertToType)

    def getTagListAndNames(self, tagList, seriesID=0, instanceID=0):
        return self[seriesID].getTagListAndNames(tagList, dsID=instanceID)

    def setTags_all(self, tag, value):
        for i in self:
            i.setTags_all(tag, value)

    def isCompressed(self):
        return self[0].isCompressed()

    def getNumberOfDicoms(self):
        return sum([len(i) for i in self])

    def getTopDir(self):
        return os.path.split(os.path.split(self[0][0].filename)[0])[0]

    def getSeriesMatchingDescription(self, matchingStrList, RETURN_SERIES_OBJ=False, REDUCE_MULTIPLE=False):
        matchStrList_lower = [i.lower() for i in matchingStrList]
        possibles = []
        for i in self:
            sDesc = i.getTag('SeriesDescription').lower()
            if any([j in sDesc  for j in matchStrList_lower ]):
                possibles.append(i)
        minID = None
        if len(possibles) == 0:
            return None
        elif (len(possibles) > 1) and REDUCE_MULTIPLE:
            seNums = [int(i.getTag('SeriesNumber')) for i in possibles]
            seNums = [i for i in seNums if i!=0]
            for k1 in range(1, len(seNums)):
                if seNums[k1] < seNums[minID]:
                    minID = k1
        if minID is not None:
            possibles = possibles[minID]
        if RETURN_SERIES_OBJ:
            return possibles
        return [f'SE{possibles[i].getTag("SeriesnNumber")}:{possibles[i].getTag("SeriesDescription")}' for i in range(len(possibles))]

    def getStudyOverview(self, tagList=SpydcmTK_config.STUDY_OVERVIEW_TAG_LIST):
        return self.getTagListAndNames(tagList)

    def getPatientOverview(self, tagList=SpydcmTK_config.SUBJECT_OVERVIEW_TAG_LIST):
        return self.getTagListAndNames(tagList)

    def getSeriesByID(self, ID):
        ID = int(ID)
        for iSeries in self:
            try:
                if int(iSeries.getTag('SeriesNumber')) == ID:
                    return iSeries
            except TypeError: # If SE# == None (report of Sec Capture etc)
                pass

    def getSeriesBySeriesNumber(self, seNum):
        return self.getSeriesByID(seNum)

    def getSeriesByUID(self, UID):
        for iSeries in self:
            if iSeries.getTag('SeriesInstanceUID') == UID:
                return iSeries


    def getSeriesByTag(self, tag, value, convertToType=None):
        for iSeries in self:
            if iSeries.getTag(tag, convertToType=convertToType) == value:
                return iSeries


    def mergeSeriesVolumesWithTime(self):
        nTimes = len(self)
        triggerTimes = sorted([self.getTag('TriggerTime', seriesID=i, instanceID=0, ifNotFound=0, convertToType=float) for i in range(nTimes)])
        nPerTime = len(self.getSeriesByTag('TriggerTime', triggerTimes[0], convertToType=float))
        sorted_ds_list = [None for _ in range(nTimes*nPerTime)]
        print(f"Merging series: n times: {nTimes}, N per time: {nPerTime}")
        for i in range(nTimes):
            iSeries = self.getSeriesByTag('TriggerTime', triggerTimes[i], convertToType=float)
            iSeries.sortByInstanceNumber()
            for k1, iDS in enumerate(iSeries):
                if nPerTime > 1:
                    # thisID = int(iDS['InstanceNumber'].value) + c1*(len(triggerTimes)-1) + c0 -1 # change to 0 index
                    thisID = i + (nTimes * k1)
                else:
                    thisID = i
                iDS.InstanceNumber = thisID + 1 # 1 index
                sorted_ds_list[thisID] = iDS
        if sorted_ds_list.count(None) != 0:
            raise ValueError('Missing some volumes')
        return DicomSeries(sorted_ds_list)


    def writeFDQ(self, seriesNumber_list, outputFileName, phase_factors=None, phase_offsets=None, working_dir=None, VERBOSE=True):
        """
        Writes a 4D-flow output as VTS format.

        Args:
            seriesNumber_list (list): A list of series numbers identifying mag and 3 phase series. Mag, X, Y, Z phases
            outputFileName (str): The name of the pvd file to write to.

        Returns:
            str: the pvd file written
        """
        # Check phase factors and offsets:
        if phase_factors is not None:
            if len(phase_factors) != 3:
                raise ValueError(f"phase_factors should be length 3")
        else:
            if self.getSeriesBySeriesNumber(seriesNumber_list[1]).IS_GE():
                phase_factors = [1.0, 1.0, -1.0]
                phase_offsets = [0.0, 0.0, 0.0]
            elif self.getSeriesBySeriesNumber(seriesNumber_list[1]).IS_SIEMENS():
                venc =  self.getSeriesBySeriesNumber(seriesNumber_list[1]).getVENC()
                c = -1.0 * venc
                m = (venc * 2.0) / 4096.0 # TODO: Check this
                phase_factors = [m, m, m]
                phase_offsets = [c, c, c]
            elif self.getSeriesBySeriesNumber(seriesNumber_list[1]).IS_PHILIPS(): # TODO: Check this
                phase_factors = [1.0, 1.0, 1.0]
                phase_offsets = [0.0, 0.0, 0.0]
            else:
                raise ValueError(f"Unknown scanner type")
        if phase_offsets is not None:
            if len(phase_offsets) != 3:
                raise ValueError(f"phase_offsets should be length 3")
        else:
            phase_offsets = [0.0, 0.0, 0.0]
        
        if working_dir is None:
            working_dir = os.path.dirname(outputFileName)
        intermediate_pvds = {}
        
        # Create process pool and process all directories in parallel
        origHIDE = self.HIDE_PROGRESSBAR
        self.HIDE_PROGRESSBAR = True
        with Pool() as pool:
            args = [(self.getSeriesBySeriesNumber(iSeriesNum), label, working_dir)
                    for iSeriesNum, label in zip(seriesNumber_list, ["MAG", "PX", "PY", "PZ"])]
            
            results = pool.map(_process_series_for_fdq, args)
        
        # Collect results into intermediate_pvds dictionary
        intermediate_pvds = dict(results)
        
        if VERBOSE: print(f"Combining to final 4D-flow PVD")
        fOut = dcmVTKTK.mergePhaseSeries4D(intermediate_pvds["MAG"], 
                                        [intermediate_pvds["PX"],
                                        intermediate_pvds["PY"],
                                        intermediate_pvds["PZ"]], 
                                        outputFileName, 
                                        phase_factors=phase_factors,
                                        phase_offsets=phase_offsets,
                                        scale_factor=0.001,
                                        DEL_ORIGINAL=True)
        self.HIDE_PROGRESSBAR = origHIDE
        if VERBOSE: print(f"Written {fOut}")
        return fOut


    def getStudySummaryDict(self, FORCE_STRING_KEYS=False):
        pt,pv = self.getPatientOverview()
        studySummaryDict = dict(zip(pt, pv))
        st,sv = self.getStudyOverview()
        stdyDict = dict(zip(st, sv))
        studySummaryDict.update(stdyDict)
        listSerDict = []
        for i in self:
            szt, szv = i.getSeriesOverview()
            listSerDict.append(dict(zip(szt, szv)))
        studySummaryDict['Series'] = listSerDict
        return studySummaryDict

    def getStudySummary(self, FULL=True):
        pt,pv = self.getPatientOverview()
        patStr = ','.join([' %s:%s'%(str(i), str(j)) for i,j in zip(pt,pv)])
        st,sv = self.getStudyOverview()
        studyStr = ','.join([' %s:%s'%(str(i), str(j)) for i,j in zip(st,sv)])
        if FULL:
            sst,_ = self[0].getSeriesOverview()
            seriesHeader = ','.join([str(i) for i in sst])
            seriesStr = [','.join([str(i2) for i2 in i.getSeriesOverview()[1]]) for i in self]
            try:
                seriesStr = sorted(seriesStr, key=lambda i:int(i.split(',')[0]))
            except ValueError:
                pass # Try to sort, but if fail (due to return not int somehow, then no sort)
            seriesStr_ = '\n\n'+seriesHeader+'\n    '+'\n    '.join(seriesStr)
            strOut = f"SUBJECT: {patStr}"
            strOut += f"\nSTUDY: {studyStr}"
            strOut += seriesStr_
            return strOut
        else:
            strOut = f"SUBJECT: {patStr}"
            strOut += f"\nSTUDY: {studyStr}"
            return strOut

    def getTagValuesList(self, tagList, RETURN_STR=False):
        output = []
        for i in self:
            output += i.getTagValuesList(tagList, False)
        if RETURN_STR:
            return dcmTools._tagValuesListToString(tagList, output)
        return output

    def writeToOrganisedFileStructure(self, patientOutputDir, studyPrefix=''):
        self.checkIfShouldUse_SAFE_NAMING()
        try:
            studyOutputDir = self[0].getStudyOutputDir(patientOutputDir, studyPrefix)
        except IndexError: # Study is empty
            return None
        studyOutputDirTemp = studyOutputDir+".WORKING"
        if os.path.isdir(studyOutputDir):
            os.rename(studyOutputDir, studyOutputDirTemp)
        for iSeries in self:
            iSeries.writeToOrganisedFileStructure(studyOutputDirTemp, 
                                                  SAFE_NAMING_CHECK=False)
        os.rename(studyOutputDirTemp, studyOutputDir)
        return studyOutputDir

    def writeToZipArchive(self, patientOutputDir, CLEAN_UP=True):
        studyOutputDir = self.writeToOrganisedFileStructure(patientOutputDir)
        if studyOutputDir is None:
            return
        r, f = os.path.split(studyOutputDir)
        shutil.make_archive(studyOutputDir, 'zip', os.path.join(r, f))
        fileOut = studyOutputDir+'.zip'
        if CLEAN_UP:
            shutil.rmtree(studyOutputDir)
        return fileOut

    def checkIfShouldUse_SAFE_NAMING(self):
        se_instance_set = set()
        for i in self:
            i.checkIfShouldUse_SAFE_NAMING(se_instance_set)


class ListOfDicomStudies(list):
    """
    Extends list of DicomStudies objects (creating list of list of list of pydicom.dataset)
    """
    def __init__(self, dStudiesList, OVERVIEW=False, HIDE_PROGRESSBAR=False):
        """
        Set OVERVIEW = False to read pixel data as well (at a cost)
        """
        self.OVERVIEW = OVERVIEW
        self.HIDE_PROGRESSBAR = HIDE_PROGRESSBAR
        list.__init__(self, dStudiesList)

    @classmethod
    def setFromInput(cls, input, OVERVIEW=False, HIDE_PROGRESSBAR=False, FORCE_READ=False, ONE_FILE_PER_DIR=False):
        if os.path.isdir(input):
            return ListOfDicomStudies.setFromDirectory(input, OVERVIEW=OVERVIEW, 
                                                                HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, 
                                                                FORCE_READ=FORCE_READ, 
                                                                ONE_FILE_PER_DIR=ONE_FILE_PER_DIR)
        if os.path.isfile(input):
            if input.endswith('tar'):
                return ListOfDicomStudies.setFromTar(input, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, 
                                                            FORCE_READ=FORCE_READ)
            elif input.endswith('tar.gz'):
                return ListOfDicomStudies.setFromTar(input, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, 
                                                            FORCE_READ=FORCE_READ)
            elif input.endswith('zip'):
                return ListOfDicomStudies.setFromZip(input, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, 
                                                            FORCE_READ=FORCE_READ)
            else:
                dcmDict = {}
                try:
                    dcmTools.readDicomFile_intoDict(input, dcmDict, FORCE_READ=FORCE_READ, OVERVIEW=OVERVIEW)
                    return ListOfDicomStudies.setFromDcmDict(dcmDict, OVERVIEW, HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)
                except dicom.filereader.InvalidDicomError:
                    raise IOError("ERROR READING DICOMS: SPDCMTK capable to read dicom files from directory, zip, tar or tar.gz\n")

    @classmethod
    def setFromDirectory(cls, dirName, OVERVIEW=False, HIDE_PROGRESSBAR=False, FORCE_READ=False, ONE_FILE_PER_DIR=False, extn_filter=None):
        dicomDict = dcmTools.organiseDicomHeirarchyByUIDs(dirName, 
                                                         HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, 
                                                         FORCE_READ=FORCE_READ, 
                                                         ONE_FILE_PER_DIR=ONE_FILE_PER_DIR, 
                                                         OVERVIEW=OVERVIEW,
                                                         extn_filter=extn_filter)
        obj = ListOfDicomStudies.setFromDcmDict(dicomDict, OVERVIEW, HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)
        for iStudy in obj:
            for iSeries in iStudy:
                iSeries.NOT_FULLY_LOADED = ONE_FILE_PER_DIR
        return obj

    @classmethod
    def setFromDcmDict(cls, dicomDict, OVERVIEW=False, HIDE_PROGRESSBAR=False, FORCE_READ=False):
        dStudiesList = []
        for studyUID in dicomDict.keys():
            dSeriesList = []
            for seriesUID in dicomDict[studyUID].keys():
                dSeriesList.append(DicomSeries(dicomDict[studyUID][seriesUID], OVERVIEW=OVERVIEW, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ))
            dStudiesList.append(DicomStudy(dSeriesList, OVERVIEW, HIDE_PROGRESSBAR))
        return cls(dStudiesList, OVERVIEW, HIDE_PROGRESSBAR)

    @classmethod
    def setFromTar(cls, tarFileName, HIDE_PROGRESSBAR=False, FORCE_READ=False, FIRST_ONLY=False, matchTagPair=None):
        # Note need OVERVIEW = False as only get access to file (and pixels) on untaring (maybe only if tar.gz) 
        dicomDict = dcmTools.getDicomDictFromTar(tarFileName, FORCE_READ=FORCE_READ, FIRST_ONLY=FIRST_ONLY, OVERVIEW_ONLY=False,
                                    matchingTagValuePair=matchTagPair, QUIET=True)
        return ListOfDicomStudies.setFromDcmDict(dicomDict, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)

    @classmethod
    def setFromZip(cls, zipFileName, HIDE_PROGRESSBAR=False, FORCE_READ=False, FIRST_ONLY=False, matchTagPair=None):
        dicomDict = dcmTools.getDicomDictFromZip(zipFileName, FORCE_READ=FORCE_READ, FIRST_ONLY=FIRST_ONLY, OVERVIEW_ONLY=False,
                                    matchingTagValuePair=matchTagPair, QUIET=True)
        return ListOfDicomStudies.setFromDcmDict(dicomDict, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR, FORCE_READ=FORCE_READ)


    def __str__(self):
        return '%d studies, %d dicoms'%(len(self), self.getNumberOfDicoms())

    def setSafeNameMode(self):
        for iStudy in self:
            for iSeries in iStudy:
                iSeries.SAFE_NAME_MODE = True

    def resetUIDs(self):
        for i in self:
            i.resetUIDs()

    def anonymise(self, anonName, anonPatientID, anonBirthdate=True, removePrivateTags=True):
        for i in self:
            i.anonymise(anonName, anonPatientID, anonBirthdate, removePrivateTags)

    def isCompressed(self):
        return self[0].isCompressed()

    def getNumberOfDicoms(self):
        return sum([i.getNumberOfDicoms() for i in self])

    def filterByTag(self, tagName, tagValue):
        return ListOfDicomStudies([i for i in self if i.getTag(tagName) == tagValue], 
                                  OVERVIEW=self.OVERVIEW, 
                                  HIDE_PROGRESSBAR=self.HIDE_PROGRESSBAR)

    def writeToOrganisedFileStructure(self, outputRootDir):
        outDirs = []
        for iStudy in self:
            suffix = ''
            for iTag in SpydcmTK_config.SUBJECT_NAMING_TAG_LIST:
                iVal = iStudy.getTag(iTag, ifNotFound='', convertToType=str)
                if len(iVal) > 0:
                    suffix += '_'+iVal
            if len(suffix) > 0:
                ioutputRootDir = os.path.join(outputRootDir, dcmTools.cleanString(suffix))
            else:
                ioutputRootDir = outputRootDir

            ooD = iStudy.writeToOrganisedFileStructure(ioutputRootDir)
            if ooD is not None:
                outDirs.append(ooD)
        return outDirs

    def writeToZipArchive(self, outputRootDir, CLEAN_UP=True):
        outDirs = []
        for iStudy in self:
            ooD = iStudy.writeToZipArchive(outputRootDir, CLEAN_UP=CLEAN_UP)
            outDirs.append(ooD)
        return outDirs

    def getSummaryString(self, FULL=True):
        output = [i.getStudySummary(FULL) for i in self]
        return '\n\n'.join(output)

    def getTableOfTagValues(self, DICOM_TAGS, RETURN_STR):
        output = []
        for i in self:
            output += i.getTagValuesList(DICOM_TAGS, False)
        if RETURN_STR:
            return dcmTools._tagValuesListToString(DICOM_TAGS, output)
        return output

    def printSummaryTable(self):
        overviewList = self.getTableOfTagValues(SpydcmTK_config.SUBJECT_OVERVIEW_TAG_LIST+SpydcmTK_config.STUDY_OVERVIEW_TAG_LIST, False)
        print(','.join(SpydcmTK_config.SUBJECT_OVERVIEW_TAG_LIST+SpydcmTK_config.STUDY_OVERVIEW_TAG_LIST))
        for i in overviewList:
            print(','.join(i))

    def getStudyByDate(self, date_str):
        return self.getStudyByTag('StudyDate', date_str)

    def getStudyByPID(self, PID):
        return self.getStudyByTag('PatientID', PID)

    def getStudyByTag(self, tagName, tagValue):
        for i in self:
            if i.getTag(tagName) == tagValue:
                return i

    def buildMSTable(self, DICOM_TAGS=SpydcmTK_config.SERIES_OVERVIEW_TAG_LIST):
        pass
        #TODO - need to pass series name or something to query and then calc mean / stdev etc

## =====================================================================================================================
##   BIDS
## =====================================================================================================================
class BIDSConst(object):
    anat = 'anat'
    func = 'func'
    dwi = 'dwi'
    fmap = 'fmap'
    code = 'code'
    configf = 'dcm2bids_config.json'


def getJsonDict(niiORjsonFile):
    if niiORjsonFile.endswith('.nii.gz'):
        niiORjsonFile = niiORjsonFile.replace('.nii.gz','.json')
    with open(niiORjsonFile, 'r') as f:
        jsonDict = json.load(f)
    return jsonDict

def getTagFromJson(niiORjsonFile, tag):
    jsonDict = getJsonDict(niiORjsonFile)
    return jsonDict[tag]


class BIDS_Subject(object):
    """
    Class to work with a single BIDS subject
    """
    def __init__(self, subjID, BIDSDirObj):
        if subjID.startswith('sub-'):
            self.subjID = subjID
        else:
            self.subjID = 'sub-%s'%(subjID)
        self.parent = BIDSDirObj

    def __gt__(self, other):
        return self.subjID > other.subjID

    def __eq__(self, other):
        return self.subjID == other.subjID

    def getSubjID(self):
        return self.subjID.replace('sub-', '')

    def getSessionDir(self, session):
        if not session.startswith('ses-'):
            session = 'ses-%s'%(session)
        return os.path.join(self.parent.rootDir, self.subjID, session)

    def getTypeDir(self, TYPE, session):
        return os.path.join(self.getSessionDir(session), TYPE)

    def _getAllTypeNII(self, TYPE, session):
        dd = self.getTypeDir(TYPE, session)
        return [os.path.join(dd, i) for i in os.listdir(dd) if i.endswith('nii.gz')]

    def getNIIMatchingModality(self, modality, TYPE, session):
        matchingF = [i for i in self._getAllTypeNII(TYPE, session) if i.endswith('_%s.nii.gz'%(modality))]
        return matchingF

    def getNIIMatchingModality_condition(self, modality, TYPE, session, tag, func):
        matchingNII = self.getNIIMatchingModality(modality, TYPE, session)
        if len(matchingNII) == 0:
            return None
        elif len(matchingNII) == 1:
            return matchingNII[0]
        else: # more than 1 run so need to make decision
            tagList = [getTagFromJson(i, tag) for i in matchingNII]
            ID = func(tagList)
            return matchingNII[ID]


class BIDS_Directory(object):
    """
    Class to work with BIDS directory
    """
    def __init__(self, rootDirectroy):
        self.rootDir = rootDirectroy

    def getConfigFile(self, configFName=BIDSConst.configf):
        return os.path.join(self.rootDir, BIDSConst.code, configFName)

    def getSubjList(self):
        subjIDs = [i for i in os.listdir(self.rootDir) if i.startswith('sub-')]
        return sorted([BIDS_Subject(i, self) for i in subjIDs])


def organiseDicoms(dcmDirectory, outputDirectory, anonName=None, anonID='', FORCE_READ=False, 
                   HIDE_PROGRESSBAR=False, REMOVE_PRIVATE_TAGS=False):
    """ Recurse down directory tree - grab dicoms and move to new
        hierarchical folder structure
    """
    if not HIDE_PROGRESSBAR:
        print('READING...')
    studies = ListOfDicomStudies.setFromInput(dcmDirectory, FORCE_READ=FORCE_READ, OVERVIEW=False, HIDE_PROGRESSBAR=HIDE_PROGRESSBAR)
    if not HIDE_PROGRESSBAR:
        print('WRITTING...')
    res = studies.writeToOrganisedFileStructure(outputDirectory, anonName=anonName, anonID=anonID, REMOVE_PRIVATE_TAGS=REMOVE_PRIVATE_TAGS)
    if res is None:
        print('    ## WARNING - No valid dicoms found')


def studySummary(pathToDicoms):
    """
    :param pathToDicoms: path to dicoms - anywhere in the tree above - will search down tree till find a dicom
    :return: a formatted string
    """
    dStudies = ListOfDicomStudies.setFromDirectory(pathToDicoms)
    return dStudies.getSummaryString()



## =====================================================================================================================
##   WRITE DICOMS
## =====================================================================================================================
def writeVTIToDicoms(vtiFile, dcmTemplateFile_or_ds, outputDir, arrayName=None, tagUpdateDict=None, patientMeta=None):
    if type(vtiFile) == str:
        vti = dcmVTKTK.fIO.readVTKFile(vtiFile)
        if vtiFile.endswith(".nii") or vtiFile.endswith(".nii.gz"):
            # TODO - need to also adjust patientMeta if this is nifti 
            vti.SetOrigin([i*0.001 for i in vti.GetOrigin()])
            vti.SetSpacing([i*0.001 for i in vti.GetSpacing()])
    else:
        vti = vtiFile
    if arrayName is None:
        A = dcmVTKTK.vtkfilters.getScalarsAsNumpy(vti)
    else:
        A = dcmVTKTK.vtkfilters.getArrayAsNumpy(vti, arrayName)
    A = np.reshape(A, vti.GetDimensions(), 'F')
    if patientMeta is None:
        patientMeta = dcmVTKTK.PatientMeta()
    patientMeta.initFromVTI(vti)
    return writeNumpyArrayToDicom(A, dcmTemplateFile_or_ds, patientMeta, outputDir, tagUpdateDict=tagUpdateDict)


def writeNumpyArrayToDicom(pixelArray, dcmTemplate_or_ds, patientMeta, outputDir, tagUpdateDict=None):
    if tagUpdateDict is None:
        tagUpdateDict = {}
    if dcmTemplate_or_ds is None:
        dsRAW = dcmTools.buildFakeDS()
    elif type(dcmTemplate_or_ds) == str:
        dsRAW = dicom.dcmread(dcmTemplate_or_ds)
    else:
        dsRAW = dcmTemplate_or_ds
    nRow, nCol, nSlice = pixelArray.shape
    if pixelArray.dtype != np.int16:
        pixelArray = pixelArray * (2**13 / np.max(pixelArray) )
        pixelArray = pixelArray.astype(np.int16)
    mx, mn = np.max(pixelArray), 0
    try:
        slice0 = tagUpdateDict.pop('SliceLocation0')
    except KeyError:
        slice0 = patientMeta.SliceLocation0*dcmVTKTK.m_to_mm
    
    sliceThick = patientMeta.SliceThickness*dcmVTKTK.m_to_mm
    ipp = [i*dcmVTKTK.m_to_mm for i in patientMeta.ImagePositionPatient]

    SeriesUID = dicom.uid.generate_uid()
    try:
        SeriesNumber = tagUpdateDict.pop('SeriesNumber')
    except KeyError:
        try:
            SeriesNumber = dsRAW.SeriesNumber * 100
        except AttributeError:
            SeriesNumber = 99
    dsList = []
    for k in range(nSlice):
        ds = copy.deepcopy(dsRAW)
        ds.SeriesInstanceUID = SeriesUID
        ds.SOPInstanceUID = dicom.uid.generate_uid()
        # ds.SpecificCharacterSet = 'ISO_IR 100'
        # ds.SOPClassUID = 'SecondaryCaptureImageStorage'
        ds.Rows = nRow
        ds.Columns = nCol
        ds.ImagesInAcquisition = nSlice
        ds.InStackPositionNumber = k+1
        # ds.RawDataRunNumber = k+1
        ds.SeriesNumber = SeriesNumber
        ds.InstanceNumber = k+1
        ds.SamplesPerPixel = 1
        ds.BitsAllocated = 16
        ds.BitsStored = 16
        ds.HighBit = 15
        ds.SliceThickness = sliceThick # Can no longer claim overlapping slices if have modified
        ds.SpacingBetweenSlices = sliceThick
        ds.SmallestImagePixelValue = int(mn)
        ds.LargestImagePixelValue = int(mx)
        ds.WindowCenter = int(mx / 2)
        ds.WindowWidth = int(mx / 2)
        ds.PixelSpacing = [i*dcmVTKTK.m_to_mm for i in list(patientMeta.PixelSpacing)]

        sliceVec = np.array(patientMeta.SliceVector)
        ImagePositionPatient = ipp + k*sliceVec*sliceThick
        ds.ImagePositionPatient = list(ImagePositionPatient)
        sliceLoc = slice0 + k*sliceThick
        ds.SliceLocation = sliceLoc
        ds.ImageOrientationPatient = list(patientMeta.ImageOrientationPatient)
        for iKey in tagUpdateDict.keys():
            if len(tagUpdateDict[iKey]) == 3:
                ds.add_new(tagUpdateDict[iKey][0], tagUpdateDict[iKey][1], tagUpdateDict[iKey][2])
            else:
                ds[iKey] = tagUpdateDict[iKey]
        ds.PixelData = pixelArray[:,:,k].tobytes()
        ds['PixelData'].VR = 'OW'
        dsList.append(ds)
    dcmSeries = DicomSeries(dsList, HIDE_PROGRESSBAR=True)
    return dcmSeries.writeToOrganisedFileStructure(outputDir)

def writeImageStackToDicom(images_sortedList, patientMeta, dcmTemplateFile_or_ds, 
                            outputDir, tagUpdateDict=None):

    combinedImage = dcmVTKTK.readImageStackToVTI(images_sortedList, patientMeta, CONVERT_TO_GREYSCALE=True)
    writeVTIToDicoms(combinedImage, 
                        dcmTemplateFile_or_ds=dcmTemplateFile_or_ds, 
                        outputDir=outputDir,
                        arrayName='PixelData',
                        tagUpdateDict=tagUpdateDict,
                        patientMeta=patientMeta)


def getResolution(dataVts):
    o,p1,p2,p3 = [0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]
    i0,i1,j0,j1,k0,k1 = dataVts.GetExtent()
    dataVts.GetPoint(i0,j0,k0, o)
    dataVts.GetPoint(i0+1,j0,k0, p1)
    dataVts.GetPoint(i0,j0+1,k0, p2)
    dataVts.GetPoint(i0,j0,k0+1, p3)
    di = abs(distBetweenTwoPts(p1, o))
    dj = abs(distBetweenTwoPts(p2, o))
    dk = abs(distBetweenTwoPts(p3, o))
    return [di, dj, dk]

def distBetweenTwoPts(a, b):
    return np.sqrt(np.sum((np.asarray(a) - np.asarray(b)) ** 2))

def _process_series_for_fdq(args):
    series, label, outDir = args
    series._loadToMemory()
    iOut = series.writeToVTS(os.path.join(outDir, f"{label}.pvd"))
    return (label, iOut)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Note: To use the 'upload' functionality of this file, you must:
#   $ pipenv install twine --dev

import io
import os

from setuptools import find_packages, setup

# Package meta-data.
NAME = 'spydcmtk'
DESCRIPTION = 'Simple python dicom toolkit.'
URL = 'https://github.com/fraser29/spydcmtk'
EMAIL = 'callaghan.fm@gmail.com'
AUTHOR = 'Fraser M. Callaghan'
REQUIRES_PYTHON = '>=3.9.0'
VERSION = '1.2.5'
KEYWORDS="medical, imaging, mri, ct, dicom"

# What packages are required for this module to be executed?
REQUIRED = [
    'pydicom>=3.0.1', 'numpy>=2.2.3', 'tqdm>=4.66.1', 'vtk>=9.3.0', 'python-gdcm', 'highdicom==0.25.0', 'matplotlib', 'ngawari>=0.1.2'
]


here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
    with open(os.path.join(here, project_slug, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION



# Where the magic happens:
setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    keywords=KEYWORDS,
    packages=find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    # If your package is a single module, use this instead of 'packages':
    # py_modules=['mypackage'],
    entry_points={
        "console_scripts": ["spydcmtk=spydcmtk.spydcm:main"],
    },
    package_data={NAME: ['spydcmtk.conf', 'ParaViewGlance.html']},
    install_requires=REQUIRED,
    # extras_require=EXTRAS,
    include_package_data=True,
    license='MIT',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        "Topic :: Scientific/Engineering :: Image Processing",
    ],
)



# -*- coding: utf-8 -*-

import configparser
import json
from collections import OrderedDict
import os


thisConfFileName = 'spydcmtk.conf'
rootDir = os.path.abspath(os.path.dirname(__file__))


class _SpydcmTK_config():
    def __init__(self, ) -> None:
        

        self.config = configparser.ConfigParser(dict_type=OrderedDict)
        self.all_config_files = [os.path.join(rootDir,thisConfFileName), 
                            os.path.join(os.path.expanduser("~"),thisConfFileName),
                            os.path.join(os.path.expanduser("~"),'.'+thisConfFileName), 
                            os.path.join(os.path.expanduser("~"), '.config',thisConfFileName),
                            os.environ.get("SPYDCMTK_CONF", '')]

    def runconfigParser(self, extraConfFile=None):
        
        if extraConfFile is not None:
            if os.path.isfile(extraConfFile):
                self.all_config_files.append(extraConfFile)
            else:
                print(f"WARNING: {extraConfFile} passed as config file to read, but FileNotFound - skipping")

        self.config.read(self.all_config_files)

        self.environment = self.config.get("app", "environment")
        self.DEBUG = self.config.getboolean("app", "debug", fallback=False)
        self.dcm2nii_path = self.config.get("app", "dcm2nii_path")
        self.dcm2nii_options = self.config.get("app", "dcm2nii_options", fallback='')

        # default_items = [vv for _,vv in self.config.items('DEFAULT')]
        extra_error_info = "Ensure to use double not single quotes for strings. "

        ## READ TAGS LISTS FOR INFO / OUTPUT NAMING
        ConfigGroupList = ["SERIES_OVERVIEW_TAG_LIST", 
                           "STUDY_OVERVIEW_TAG_LIST",
                           "SUBJECT_OVERVIEW_TAG_LIST",
                           "MANUSCRIPT_TABLE_EXTRA_TAG_LIST",
                           "VTI_NAMING_TAG_LIST",
                           "SUBJECT_NAMING_TAG_LIST",
                           "STUDY_NAMING_TAG_LIST",
                           "SERIES_NAMING_TAG_LIST"]
        for iGroup in ConfigGroupList:

            try:
                setattr(self, iGroup, json.loads(self.config.get(iGroup.lower(), "tagList")))
            except json.decoder.JSONDecodeError:
                print(f"Error reading {iGroup} from spydcmtk.conf.")
                print(f"  {extra_error_info}")

    def printInfo(self):
        print(" ----- SPYDCMTK Configuration INFO -----")
        print('   Using configuration files found at: ')
        for iFile in self.all_config_files:
            if os.path.isfile(iFile):
                print(f"    {iFile}")
        print('')
        print('   Configuration settings:')
        attributes = vars(self)
        for attribute_name in sorted(attributes.keys()):
            if 'config' in attribute_name:
                continue
            print(f"   --  {attribute_name}: {attributes[attribute_name]}")


SpydcmTK_config = _SpydcmTK_config()
SpydcmTK_config.runconfigParser()